"""
SHT31 config file.
"""
# built-in imports
import munch

# local imports
from thermostatsupervisor import utilities as util

ALIAS = "sht31"

# SHT31 thermometer zones
LOFT_SHT31 = 0  # zone 0, local IP 192.168.86.15
LOFT_SHT31_REMOTE = 1  # zone 1

# unit test parameters
UNIT_TEST_ZONE = 99
UNIT_TEST_SEED = 0x7F
UNIT_TEST_ENV_KEY = "SHT31_REMOTE_IP_ADDRESS_" + str(UNIT_TEST_ZONE)
FLASK_PORT = 5000  # note: ports below 1024 require root access on Linux
FLASK_USE_HTTPS = False  # HTTPS requires a cert to be installed.
FLASK_DEBUG_MODE = False  # True to enable flask debugging mode
if FLASK_USE_HTTPS:
    FLASK_SSL_CERT = 'adhoc'  # adhoc
    FLASK_KWARGS = {'ssl_context': FLASK_SSL_CERT}
    FLASK_URL_PREFIX = "https://"
else:
    FLASK_SSL_CERT = None  # adhoc
    FLASK_KWARGS = {}
    FLASK_URL_PREFIX = "http://"

# diagnostic parameters
flask_folder = munch.Munch()
flask_folder.production = ""
flask_folder.unit_test = "/unit"
flask_folder.diag = "/diag"
flask_folder.clear_diag = "/clear_diag"
flask_folder.enable_heater = "/enable_heater"
flask_folder.disable_heater = "/disable_heater"
flask_folder.soft_reset = "/soft_reset"
flask_folder.reset = "/reset"
flask_folder.i2c_recovery = "/i2c_recovery"
flask_folder.i2c_detect = "/i2c_detect"
flask_folder.i2c_detect_0 = "/i2c_detect/0"
flask_folder.i2c_detect_1 = "/i2c_detect/1"

# SHT31 API field names
API_MEASUREMENT_CNT = 'measurements'
API_TEMPC_MEAN = 'Temp(C) mean'
API_TEMPC_STD = 'Temp(C) std'
API_TEMPF_MEAN = 'Temp(F) mean'
API_TEMPF_STD = 'Temp(F) std'
API_HUMIDITY_MEAN = 'Humidity(%RH) mean'
API_HUMIDITY_STD = 'Humidity(%RH) std'

# SHT31D config
I2C_BUS = 1  # 0 = /dev/i2c-0 (port I2C0), 1 = /dev/i2c-1 (port I2C1)
I2C_ADDRESS = 0x45  # i2c address, 0x44=a, 0x45=b
MEASUREMENTS = 10  # number of MEASUREMENTS to average

# pi0 / sht31 connection config, -1 means non-addressible pin
V3_PIN = -1  # 3.3v power pin (red), (pi pin 1)
SDA_PIN = 2  # i2c data signal (brown), GPIO2 (pi pin 3)
SCL_PIN = 3  # i2c clock signal (orange), GPIO3 (pi pin 5)
ADDR_PIN = 4  # i2c address (white), GPIO4, low = 0x44, high=0x45 (pi pin 7)
GND_PIN = -1  # ground wire (black), (pi pin 9)
ALERT_PIN = 17  # i2c alert pint (yellow), GPIO17 (pi pin 11)

# all environment variables specific to this thermostat type
env_variables = {
    "SHT31_REMOTE_IP_ADDRESS_0": None,
    "SHT31_REMOTE_IP_ADDRESS_1": None,
    UNIT_TEST_ENV_KEY: None,
}

# min required env variables on all runs
required_env_variables = {
    "SHT31_REMOTE_IP_ADDRESS_": None,  # prefix only, excludes zone
}

# supported thermostat configs
supported_configs = {"module": "sht31",
                     "type": 3,
                     "zones": [LOFT_SHT31, LOFT_SHT31_REMOTE, UNIT_TEST_ZONE],
                     "modes": ["OFF_MODE", "UNKNOWN_MODE"]}

# metadata dict:
# 'zone_name' is returned by self.get_zone_name()
# 'host_name' is used for DNS lookup to determine if device
# is on the current network.
metadata = {
    LOFT_SHT31: {"zone_name": "Loft (local)",
                 "host_name": "raspberrypi0.lan",
                 },
    LOFT_SHT31_REMOTE: {"zone_name": "loft (remote)",
                        "host_name": "none",
                        },
    UNIT_TEST_ZONE: {"zone_name": "unittest",
                     "host_name": "none",
                     },
}


def get_preferred_zone():
    """
    Return the preferred zone number.  For this thermostat the preferred zone
    number is the local zone if present, otherwise will fail over the the
    remote zone.

    inputs:
        None
    returns:
        (int): zone number.
    """
    # loopback does not work so use local sht31 zone if testing
    # on the local net.  If not, use the DNS name.
    local_host = metadata[LOFT_SHT31]["host_name"]
    zone = str([LOFT_SHT31_REMOTE, LOFT_SHT31][
        util.is_host_on_local_net(local_host, verbose=False)[0]])
    return zone


default_zone = get_preferred_zone()
default_zone_name = ALIAS + "_" + str(default_zone)

argv = [
    "supervise.py",  # module
    ALIAS,  # thermostat
    str(default_zone),  # zone
    "17",  # poll time in sec
    "357",  # reconnect time in sec
    "2",  # tolerance
    "OFF_MODE",  # thermostat mode
    "2",  # number of measurements
    ]



"""
Unit test module for environment.py.
"""
# built-in imports
import os
import sys
import unittest

# local imports
import thermostatsupervisor
from thermostatsupervisor import emulator_config
from thermostatsupervisor import environment as env
from thermostatsupervisor import utilities as util
from tests import unit_test_common as utc


class EnvironmentTests(utc.UnitTest):
    """Test functions related to environment and env variables."""

    def setUp(self):
        super().setUp()
        util.log_msg.file_name = "unit_test.txt"

    def test_is_interactive_environment(self):
        """
        Verify is_interactive_environment().
        """
        return_val = env.is_interactive_environment()
        self.assertTrue(isinstance(return_val, bool))

    def test_get_env_variable(self):
        """
        Confirm get_env_variable() can retrieve values.
        """
        for env_key in ["GMAIL_USERNAME", "GMAIL_PASSWORD"]:
            buff = env.get_env_variable(env_key)
            print(f"env${env_key}="
                  f"{[buff['value'], '(hidden)']['PASSWORD' in env_key]}")
            self.assertEqual(buff["status"], util.NO_ERROR)
            self.assertGreater(len(buff["value"]), 0)

    def test_load_all_env_variables(self):
        """
        Confirm all env variables can be loaded.
        """
        env.load_all_env_variables()
        print(f"env var dict={env.env_variables}")

    def test_get_local_ip(self):
        """
        Verify get_local_ip().
        """
        return_val = env.get_local_ip()
        self.assertTrue(isinstance(return_val, str),
                        "get_local_ip() returned '%s' which is not a string")
        self.assertTrue(7 <= len(return_val) <= 15,
                        "get_local_ip() returned '%s' which is not "
                        "between 7 and 15 chars")

    def test_is_host_on_local_net(self):
        """
        Verify is_host_on_local_net() runs as expected.

        Test cases need to be site-agnostic or require some type
        of filtering to ensure they pass regardless of which LAN
        this test is running from.

        util.is_host_on_local_net is not reliable when passing
        in an IP address so most test cases are for hostname only.
        """
        test_cases = [
            # [host_name, ip_address, expected_result]
            ['testwifi.here', None,
             not env.is_azure_environment()],  # Google wifi router
            ['bogus_host', '192.168.86.145', False],  # bogus host
            ['bogus_host', None, False],  # bogus host without IP
            ['dns.google', '8.8.8.8', True],  # should pass everywhere
        ]

        for test_case in test_cases:
            print(f"testing for '{test_case[0]}' at {test_case[1]}, expect "
                  f"{test_case[2]}")
            result, ip_address = util.is_host_on_local_net(test_case[0],
                                                           test_case[1],
                                                           True)
            # verify IP length returned
            if result:
                ip_length_symbol = ">="
                ip_length_min = 7
                self.assertTrue(len(ip_address) >= ip_length_min,
                                f"ip_address returned ({ip_address}) did not "
                                f"meet length expectations ("
                                f"{ip_length_symbol + str(ip_length_min)})")
            else:
                self.assertTrue(ip_address is None,
                                f"ip_address returned ({ip_address}) "
                                f"is not None")

            # verify expected result
            self.assertEqual(result, test_case[2],
                             f"test_case={test_case[0]}, expected="
                             f"{test_case[2]}, actual={result}")

    def test_get_python_version(self):
        """Verify get_python_version()."""
        major_version, minor_version = env.get_python_version()

        # verify major version
        min_major = int(env.MIN_PYTHON_MAJOR_VERSION)
        self.assertTrue(major_version >= min_major,
                        f"python major version ({major_version}) is not gte "
                        f"min required value ({min_major})")

        # verify minor version
        min_minor = int(str(env.MIN_PYTHON_MAJOR_VERSION)[
            str(env.MIN_PYTHON_MAJOR_VERSION).find(".") + 1:])
        self.assertTrue(minor_version >= min_minor,
                        f"python minor version ({minor_version}) is not gte "
                        f"min required value ({min_minor})")

        # error checking invalid input parmeter
        with self.assertRaises(TypeError):
            print("attempting to invalid input parameter type, "
                  "expect exception...")
            env.get_python_version("3", 7)

        # no decimal point
        env.get_python_version(3, None)

        # min value exception
        with self.assertRaises(EnvironmentError):
            print("attempting to verify version gte 99.99, "
                  "expect exception...")
            env.get_python_version(99, 99)

        print("test passed all checks")

    def test_dynamic_module_import(self):
        """
        Verify dynamic_module_import() runs without error

        TODO: this module results in a resourcewarning within unittest:
        sys:1: ResourceWarning: unclosed <socket.socket fd=628,
        family=AddressFamily.AF_INET, type=SocketKind.SOCK_DGRAM, proto=0,
        laddr=('0.0.0.0', 64963)>
        """

        # test successful case
        package_name = util.PACKAGE_NAME + "." + emulator_config.ALIAS
        pkg = env.dynamic_module_import(package_name)
        print(f"default thermostat returned package type {type(pkg)}")
        self.assertTrue(isinstance(pkg, object),
                        f"dynamic_module_import() returned type({type(pkg)}),"
                        f" expected an object")
        del sys.modules[package_name]
        del pkg

        # test failing case
        with self.assertRaises(ImportError):
            print("attempting to open bogus package name, expect exception...")
            pkg = env.dynamic_module_import(util.PACKAGE_NAME + "." + "bogus")
            print(f"'bogus' module returned package type {type(pkg)}")
        print("test passed")

    def test_get_parent_path(self):
        """
        Verify get_parent_path().
        """
        return_val = env.get_parent_path(os.getcwd())
        self.assertTrue(isinstance(return_val, str),
                        "get_parent_path() returned '%s' which is not a string")

    def test_get_package_version(self):
        """
        Verify get_package_version().
        """
        pkg = thermostatsupervisor
        return_type = tuple
        return_val = env.get_package_version(pkg)
        self.assertTrue(isinstance(return_val, return_type),
                        f"return_val = {return_val}, expected type "
                        f"{return_type}, actual_type {type(return_val)}")

        # check individual elements
        elements = [
            "major",
            "minor",
            "patch",
            ]
        return_type = int
        for element in elements:
            return_val = env.get_package_version(pkg, element)
            self.assertTrue(isinstance(return_val, return_type),
                            f"element='{element}', return_val = {return_val},"
                            " expected type "
                            f"{return_type}, actual_type {type(return_val)}")


if __name__ == "__main__":
    util.log_msg.debug = True
    unittest.main(verbosity=2)



"""
Unit test module for supervise_flask_server.py.

Flask server tests currently do not work on Azure pipelines
because ports cannot be opened on shared pool.
"""
# built-in imports
import threading
import time
import unittest

# third party imports
from flask_wtf.csrf import CSRFProtect
import requests

# local imports
from thermostatsupervisor import environment as env
from thermostatsupervisor import flask_generic as flg
# thermostat_api is imported but not used to avoid a circular import
from thermostatsupervisor import thermostat_api as api  # noqa F401, pylint: disable=unused-import.
from thermostatsupervisor import supervisor_flask_server as sfs
from thermostatsupervisor import utilities as util
from tests import unit_test_common as utc


@unittest.skipIf(env.is_azure_environment(),
                 "this test not supported on Azure Pipelines")
@unittest.skipIf(not env.is_interactive_environment(),
                 "this test hangs when run from the command line")
@unittest.skipIf(not utc.ENABLE_FLASK_INTEGRATION_TESTS,
                 "flask integration tests are disabled")
class IntegrationTest(utc.UnitTest):
    """Test functions in supervisor_flask_server.py."""

    app = sfs.create_app()
    csrf = CSRFProtect(app)  # enable CSRF protection
    ip_ban = flg.initialize_ipban(app)  # hacker blacklisting agent
    flg.set_flask_cookie_config(app)
    flg.print_flask_config(app)

    def setUp(self):
        super().setUp()
        sfs.debug = False
        sfs.measurements = 10
        sfs.unit_test_mode = True
        util.log_msg.file_name = "unit_test.txt"
        if not env.is_azure_environment():
            # mock the argv list
            sfs.argv = utc.unit_test_argv
            print("starting supervise flask server thread...")
            self.flask_server = threading.Thread(
                target=sfs.app.run, args=('0.0.0.0', sfs.FLASK_PORT, False),
                kwargs=sfs.flask_kwargs)
            self.flask_server.daemon = True  # make thread daemonic
            self.flask_server.start()
            print(f"thread alive status={self.flask_server.is_alive()}")
            print("Flask server setup is complete")
        else:
            print("WARNING: flask server tests not currently supported on "
                  "Azure pipelines, doing nothing")

    def tearDown(self):
        if not env.is_azure_environment():
            print(f"thread alive status={self.flask_server.is_alive()}")
            if self.flask_server.daemon:
                print("flask server is daemon thread, "
                      "thread will terminate when main thread terminates")
            else:
                print("WARNING: flask server is not daemon thread, "
                      "thread may still be active")
        super().tearDown()

    def test_supervisor_flask_server(self):
        """
        Confirm Flask server returns valid data.

        This test requires a live thermostat connection to run the
        supervise routine on.
        """
        # grab supervise web page result and display
        flask_url = (sfs.FLASK_URL_PREFIX + env.get_local_ip() + ':' +
                     str(sfs.FLASK_PORT))

        # delay for page load and initial data posting
        wait_delay_sec = 10
        polling_interval_sec = 4
        while wait_delay_sec > 0:
            print(f"waiting {wait_delay_sec:.0f} seconds for initial "
                  "supervisor page to be populated...")
            wait_delay_sec -= polling_interval_sec
            time.sleep(polling_interval_sec)  # polling interval

        # grab web page and check response code
        print(f"grabbing web page results from: {flask_url}")
        results = requests.get(flask_url, timeout=util.HTTP_TIMEOUT)
        print(f"web page response code={results.status_code}")
        self.assertEqual(results.status_code, 200,
                         f"web page response was {results.status_code}, "
                         f"expected 200")

        # check web page content vs. expectations
        print(f"web page contents: {results.content}")
        exp_substr = (
            f"<title>{utc.unit_test_argv[1]} thermostat zone "
            f"{utc.unit_test_argv[2]}, {utc.unit_test_argv[7]} "
            f"measurements</title>")
        self.assertTrue(exp_substr in results.content.decode("utf-8"),
                        f"did not find substring '{exp_substr}' in web page"
                        " response")


if __name__ == "__main__":
    util.log_msg.debug = True
    unittest.main(verbosity=2)



"""
File to automate unit test coverage.

dependencies:
pip install coverage (one time install)

to run this module:
from command prompt:  'python unit_test_coverage.py'
or from IDE: 'Run / Run As / 1. Python Run'
note: this module will not run properly as a unit test.

coverage results:
session window will contain high level coverage report
open /htmlcov/index.html to see the html report index.
"""
# built-in imports
import coverage


def code_coverage_all_tests():
    """
    Run all enabled unit tests and collect code coverage data.
    """
    # start the coverage service
    cov = coverage.Coverage()
    cov.start()

    # run all unit tests
    # defer imports until after coverage service
    # starts so that all imports are included in
    # coverage metric.
    # pylint: disable=import-outside-toplevel
    from tests import unit_test_common as utc  # noqa E402
    try:
        utc.parse_unit_test_runtime_parameters()
        utc.run_all_tests()
    finally:
        # stop the coverage service and generate reports
        cov.stop()
        cov.report()
        cov.html_report(directory="htmlcov")


if __name__ == "__main__":
    code_coverage_all_tests()



"""
Unit test module for supervise.py.
"""
# built-in imports
import unittest

# local imports
from thermostatsupervisor import emulator_config
from thermostatsupervisor import environment as env
from thermostatsupervisor import supervise as sup
from thermostatsupervisor import utilities as util
from tests import unit_test_common as utc


class Test(utc.UnitTest):
    """Test functions in supervise.py."""

    def setUp(self):
        super().setUp()
        self.setup_mock_thermostat_zone()

    def tearDown(self):
        self.teardown_mock_thermostat_zone()
        super().tearDown()

    @unittest.skipIf(env.is_azure_environment(),
                     "this test not supported on Azure Pipelines")
    def test_supervisor(self):
        """Verify main supervisor loop."""
        sup.supervisor(emulator_config.ALIAS,
                       emulator_config.supported_configs["zones"][0])


if __name__ == "__main__":
    util.log_msg.debug = True
    unittest.main(verbosity=2)



"""Blink Camera."""
# built-in imports
import asyncio
import os
import sys
import time
import traceback

# third party imports

# local imports
from thermostatsupervisor import blink_config
from thermostatsupervisor import thermostat_api as api
from thermostatsupervisor import thermostat_common as tc
from thermostatsupervisor import environment as env
from thermostatsupervisor import utilities as util

# Blink library
BLINK_DEBUG = False  # debug uses local blink repo instead of pkg
if BLINK_DEBUG and not env.is_azure_environment():
    pkg = "blinkpy.blinkpy"
    mod_path = "..\\blinkpy"
    if env.is_interactive_environment():
        mod_path = "..\\" + mod_path
    blinkpy = env.dynamic_module_import("blinkpy.blinkpy",
                                        mod_path, pkg)
    auth = env.dynamic_module_import("blinkpy.auth",
                                     mod_path, pkg)
else:
    from blinkpy import auth  # noqa E402, from path / site packages
    from blinkpy import blinkpy  # noqa E402, from path / site packages


class ThermostatClass(blinkpy.Blink, tc.ThermostatCommon):
    """Blink Camera thermostat functions."""

    def __init__(self, zone, verbose=True):
        """
        Constructor, connect to thermostat.

        inputs:
            zone(str):  zone of thermostat.
            verbose(bool): debug flag.
        """
        # Blink server auth credentials from env vars
        self.BL_UNAME_KEY = 'BLINK_USERNAME'
        self.BL_PASSWORD_KEY = 'BLINK_PASSWORD'
        self.bl_uname = (os.environ.get(self.BL_UNAME_KEY, "<" +
                         self.BL_UNAME_KEY + "_KEY_MISSING>"))
        self.bl_pwd = (os.environ.get(
            self.BL_PASSWORD_KEY, "<" +
            self.BL_PASSWORD_KEY + "_KEY_MISSING>"))
        self.auth_dict = {"username": self.bl_uname, "password": self.bl_pwd}
        self.BL_2FA_KEY = "BLINK_2FA"
        self.bl_2fa = (os.environ.get(
            self.BL_2FA_KEY, "<" +
            self.BL_2FA_KEY + "_KEY_MISSING>"))
        self.verbose = verbose
        self.zone_number = zone

        # connect to Blink server and authenticate
        self.args = None
        self.thermostat_type = None
        self.blink = None
        if env.get_package_version(blinkpy) >= (0, 22, 0):
            asyncio.run(self.async_auth_start())
        else:
            self.auth_start()

        # get cameras
        self.camera_metadata = {}
        self.get_cameras()

        # configure zone info
        self.zone_name = self.get_zone_name()
        self.device_id = None  # initialize
        self.device_id = self.get_target_zone_id(self.zone_number)
        self.serial_number = None  # will be populated when unit is queried.

    def auth_start(self):
        """
        blinkpy < 0.22.0-compatible start
        """
        # construct the superclass
        # call both parent class __init__
        self.args = [self.bl_uname, self.bl_pwd]
        blinkpy.Blink.__init__(self, *self.args)
        tc.ThermostatCommon.__init__(self)

        # set tstat type and debug flag
        self.thermostat_type = blink_config.ALIAS

        # establish connection
        self.blink = blinkpy.Blink()
        if self.blink is None:
            print(traceback.format_exc())
            raise RuntimeError("ERROR: Blink object failed to instantiate "
                               f"for zone {self.zone_number}")

        self.blink.auth = auth.Auth(self.auth_dict, no_prompt=True)
        self.blink.start()
        try:
            self.blink.auth.send_auth_key(self.blink, self.bl_2fa)
        except AttributeError:
            error_msg = ("ERROR: Blink authentication failed for zone "
                         f"{self.zone_number}, this may be due to spamming the "
                         "blink server, please try again later.")
            banner = "*" * len(error_msg)
            print(banner)
            print(error_msg)
            print(banner)
            sys.exit(1)
        self.blink.setup_post_verify()

    async def async_auth_start(self):
        """
        blinkpy 0.22.0-compatible async start
        """
        # construct the superclass
        # call both parent class __init__
        self.args = [self.bl_uname, self.bl_pwd]
        blinkpy.Blink.__init__(self, *self.args)
        tc.ThermostatCommon.__init__(self)

        # set tstat type and debug flag
        self.thermostat_type = blink_config.ALIAS

        # establish connection
        self.blink = blinkpy.Blink()
        if self.blink is None:
            print(traceback.format_exc())
            raise RuntimeError("ERROR: Blink object failed to instantiate "
                               f"for zone {self.zone_number}")

        self.blink.auth = auth.Auth(self.auth_dict, no_prompt=True)
        await self.blink.start()
        try:
            await self.blink.auth.send_auth_key(self.blink, self.bl_2fa)
        except AttributeError:
            error_msg = ("ERROR: Blink authentication failed for zone "
                         f"{self.zone_number}, this may be due to spamming the "
                         "blink server, please try again later.")
            banner = "*" * len(error_msg)
            print(banner)
            print(error_msg)
            print(banner)
            sys.exit(1)
        await self.blink.setup_post_verify()

    def get_zone_name(self):
        """
        Return the name associated with the zone number from metadata dict.

        inputs:
            None
        returns:
            (str) zone name
        """
        return blink_config.metadata[self.zone_number]["zone_name"]

    def get_target_zone_id(self, zone=0):
        """
        Return the target zone ID.

        inputs:
            zone(int): zone number.
        returns:
            (obj): Blink object
        """
        # return the target zone object
        return zone

    def get_cameras(self):
        """
        Get the blink cameras
        """
        table_length = 20
        if self.verbose:
            print("blink camera inventory:")
            print("-" * table_length)
        for name, camera in self.blink.cameras.items():
            if self.verbose:
                print(name)
                print(camera.attributes)
            self.camera_metadata[name] = camera.attributes
        if self.verbose:
            print("-" * table_length)

    def get_all_metadata(self, zone=None, debug=False):
        """Get all thermostat meta data for device_id from local API.

        inputs:
            zone(): specified zone
            debug(bool): debug flag.
        returns:
            (dict): dictionary of meta data.
        """
        return self.get_metadata(zone, None, debug)

    def get_metadata(self, zone=None, parameter=None, debug=False):
        """Get thermostat meta data for device_id from local API.

        inputs:
            zone(): specified zone
            parameter(str): target parameter, if None will return all.
            debug(bool): debug flag.
        returns:
            (dict): dictionary of meta data.
        """
        zone_name = blink_config.metadata[self.zone_number]["zone_name"]
        if self.blink.cameras == {}:
            raise ValueError("camera list is empty when searching for camera"
                             f" {zone_name}")
        for name, camera in self.blink.cameras.items():
            # print(f"DEBUG: camera {name}: {camera.attributes}")
            if name == zone_name:
                if debug:
                    print(f"found camera {name}: {camera.attributes}")
                if parameter is None:
                    return camera.attributes
                else:
                    return camera.attributes[parameter]
        raise ValueError(f"Camera zone {zone}({zone_name}) was not found")

    def print_all_thermostat_metadata(self, zone, debug=False):
        """Print all metadata for zone to the screen.

        inputs:
            zone(int): specified zone, if None will print all zones.
            debug(bool): debug flag
        returns:
            None, prints result to screen
        """
        self.exec_print_all_thermostat_metadata(
            self.get_all_metadata, [zone, debug])


class ThermostatZone(tc.ThermostatCommonZone):
    """
    KumoCloud single zone on local network.

    Class needs to be updated for multi-zone support.
    """

    def __init__(self, Thermostat_obj, verbose=True):
        """
        Zone constructor.

        inputs:
            Thermostat(obj): Thermostat class instance.
            verbose(bool): debug flag.
        """
        # construct the superclass, requires auth setup first
        super().__init__()

        # runtime parameter defaults
        self.poll_time_sec = 10 * 60  # default to 10 minutes
        self.connection_time_sec = 8 * 60 * 60  # default to 8 hours

        # server data cache expiration parameters
        self.fetch_interval_sec = 10  # age of server data before refresh
        self.last_fetch_time = time.time() - 2 * self.fetch_interval_sec

        # switch config for this thermostat
        self.system_switch_position[tc.ThermostatCommonZone.COOL_MODE] = "cool"
        self.system_switch_position[tc.ThermostatCommonZone.HEAT_MODE] = "heat"
        self.system_switch_position[tc.ThermostatCommonZone.OFF_MODE] = "off"
        self.system_switch_position[tc.ThermostatCommonZone.DRY_MODE] = "dry"
        self.system_switch_position[tc.ThermostatCommonZone.AUTO_MODE] = "auto"
        self.system_switch_position[tc.ThermostatCommonZone.FAN_MODE] = "vent"

        # zone info
        self.verbose = verbose
        self.thermostat_type = blink_config.ALIAS
        self.device_id = Thermostat_obj.device_id
        self.Thermostat = Thermostat_obj
        self.zone_number = Thermostat_obj.zone_number
        self.zone_name = self.get_zone_name()
        self.zone_metadata = Thermostat_obj.get_metadata(zone=self.zone_number)

    def get_display_temp(self) -> float:  # used
        """
        Refresh the cached zone information and return Indoor Temp in Deg F.

        inputs:
            None
        returns:
            (float): indoor temp in deg F.
        """
        raw_temp = self.zone_metadata.get(blink_config.API_TEMPF_MEAN)
        if isinstance(raw_temp, (str, float, int)):
            raw_temp = float(raw_temp)
        return raw_temp

    def get_zone_name(self):
        """
        Return the name associated with the zone number from metadata dict.

        inputs:
            None
        returns:
            (str) zone name
        """
        return blink_config.metadata[self.zone_number]["zone_name"]

    def get_display_humidity(self) -> (float, None):
        """
        Refresh the cached zone information and return IndoorHumidity.

        inputs:
            None
        returns:
            (float, None): indoor humidity in %RH, None if not supported.
        """
        return None  # not available

    def get_is_humidity_supported(self) -> bool:
        """Return humidity sensor status."""
        return self.get_display_humidity() is not None

    def is_heat_mode(self) -> int:
        """Return the heat mode."""
        return 0  # not applicable

    def is_cool_mode(self) -> int:
        """Return the cool mode."""
        return 0  # not applicable

    def is_dry_mode(self) -> int:
        """Return the dry mode."""
        return 0  # not applicable

    def is_auto_mode(self) -> int:
        """Return the auto mode."""
        return 0  # not applicable

    def is_fan_mode(self) -> int:
        """Return the fan mode."""
        return 0  # not applicable

    def is_off_mode(self) -> int:
        """Return the off mode."""
        return 1  # always off

    def is_heating(self) -> int:
        """Return 1 if actively heating, else 0."""
        return 0  # not applicable

    def is_cooling(self) -> int:
        """Return 1 if actively cooling, else 0."""
        return 0  # not applicable

    def is_drying(self):
        """Return 1 if drying relay is active, else 0."""
        return 0  # not applicable

    def is_auto(self):
        """Return 1 if auto relay is active, else 0."""
        return 0  # not applicable

    def is_fanning(self):
        """Return 1 if fan relay is active, else 0."""
        return 0  # not applicable

    def is_power_on(self):
        """Return 1 if power relay is active, else 0."""
        return 1  # always on

    def is_fan_on(self):
        """Return 1 if fan relay is active, else 0."""
        return 0  # not applicable

    def is_defrosting(self):
        """Return 1 if defrosting is active, else 0."""
        return 0  # not applicable

    def is_standby(self):
        """Return 1 if standby is active, else 0."""
        return 0  # not applicable

    def get_system_switch_position(self) -> int:
        """ Return the thermostat mode.

        inputs:
            None
        returns:
            (int): thermostat mode, see tc.system_switch_position for details.
        """
        return self.system_switch_position[self.OFF_MODE]

    def report_heating_parameters(self, switch_position=None):
        """
        Display critical thermostat settings and reading to the screen.

        inputs:
            switch_position(int): switch position override, used for testing.
        returns:
            None
        """
        # current temp as measured by thermostat
        util.log_msg(
            f"display temp="
            f"{util.temp_value_with_units(self.get_display_temp())}",
            mode=util.BOTH_LOG,
            func_name=1)

        # get switch position
        if switch_position is None:
            switch_position = self.get_system_switch_position()

        # heating status
        if switch_position == \
                self.system_switch_position[self.HEAT_MODE]:
            util.log_msg(f"heat mode={self.is_heat_mode()}",
                         mode=util.BOTH_LOG)
            util.log_msg(
                f"heat setpoint={self.get_heat_setpoint_raw()}",
                mode=util.BOTH_LOG)
            util.log_msg(
                f"schedule heat sp={self.get_schedule_heat_sp()}",
                mode=util.BOTH_LOG)

        # cooling status
        if switch_position == \
                self.system_switch_position[self.COOL_MODE]:
            util.log_msg(f"cool mode={self.is_cool_mode()}",
                         mode=util.BOTH_LOG)
            util.log_msg(
                f"cool setpoint={self.get_cool_setpoint_raw()}",
                mode=util.BOTH_LOG)
            util.log_msg(
                f"schedule cool sp={self.get_schedule_cool_sp()}",
                mode=util.BOTH_LOG)

        # hold settings
        util.log_msg(
            f"is in vacation hold mode={self.get_is_invacation_hold_mode()}",
            mode=util.BOTH_LOG)
        util.log_msg(f"vacation hold={self.get_vacation_hold()}",
                     mode=util.BOTH_LOG)
        util.log_msg(
            f"vacation hold until time={self.get_vacation_hold_until_time()}",
            mode=util.BOTH_LOG)
        util.log_msg(
            f"temporary hold until time="
            f"{self.get_temporary_hold_until_time()}",
            mode=util.BOTH_LOG)


if __name__ == "__main__":

    # verify environment
    env.get_python_version()

    # get zone override
    api.uip = api.UserInputs(argv_list=None,
                             thermostat_type=blink_config.ALIAS)
    zone_number = api.uip.get_user_inputs(api.uip.zone_name,
                                          api.input_flds.zone)

    tc.thermostat_basic_checkout(
        blink_config.ALIAS,
        zone_number,
        ThermostatClass, ThermostatZone)

    tc.thermostat_get_all_zone_temps(blink_config.ALIAS,
                                     blink_config.supported_configs["zones"],
                                     ThermostatClass,
                                     ThermostatZone)



"""
emulator thermostat config file.
"""
ALIAS = "emulator"

# constants
MAX_HEAT_SETPOINT = 66.0  # float
MIN_COOL_SETPOINT = 78.0  # float
STARTING_MODE = "OFF_MODE"  # thermostat set mode when emulator starts
STARTING_TEMP = 72.0  # starting temperature when emulator starts
NORMAL_TEMP_VARIATION = 16.0  # reported value variation +/- this value
STARTING_HUMIDITY = 45.0  # starting humidity when emulator starts
NORMAL_HUMIDITY_VARIATION = 3.0  # reported val variation +/- this val

# all environment variables specific to this thermostat type
env_variables = {
}

# min required env variables on all runs
required_env_variables = {
}

# supported thermostat configs
supported_configs = {"module": "emulator",
                     "type": 0,
                     "zones": [0, 1],
                     "modes": ["OFF_MODE", "HEAT_MODE", "COOL_MODE",
                               "DRY_MODE", "AUTO_MODE", "UNKNOWN_MODE"]}

default_zone = supported_configs["zones"][0]
default_zone_name = ALIAS + "_" + str(default_zone)

argv = [
    "supervise.py",  # module
    ALIAS,  # thermostat
    str(default_zone),  # zone
    "19",  # poll time in sec
    "359",  # reconnect time in sec
    "3",  # tolerance
    "OFF_MODE",  # thermostat mode
    "2",  # number of measurements
    ]



"""
Integration test module for sht31.py.

This test requires connection to sht31 thermostat.
"""
# built-in imports
import unittest

# local imports
from thermostatsupervisor import sht31
from thermostatsupervisor import sht31_config
from thermostatsupervisor import utilities as util
from tests import unit_test_common as utc


@unittest.skipIf(not utc.ENABLE_SHT31_TESTS,
                 "sht31 tests are disabled")
class IntegrationTest(utc.IntegrationTest):
    """
    Test functions in sht31.py.
    """

    def setUpIntTest(self):
        self.setup_common()
        self.print_test_name()

        # argv list must be valid settings
        self.unit_test_argv = [
            "supervise.py",  # module
            "sht31",  # thermostat
            # loopback does not work so use local sht31 zone if testing
            # on the local net.  If not, use the DNS name.
            str(sht31_config.default_zone),
            "5",  # poll time in sec
            "12",  # reconnect time in sec
            "2",  # tolerance
            "UNKNOWN_MODE",  # thermostat mode, no target
            "6",  # number of measurements
        ]
        self.mod = sht31
        self.mod_config = sht31_config


@unittest.skipIf(not utc.ENABLE_SHT31_TESTS,
                 "sht31 tests are disabled")
class FunctionalIntegrationTest(IntegrationTest,
                                utc.FunctionalIntegrationTest):
    """
    Test functional performance of sht31.py.
    """

    def setUp(self):
        self.setUpIntTest()
        # test_GetMetaData input parameters
        self.metadata_field = sht31_config.API_TEMPF_MEAN
        self.metadata_type = float


@unittest.skipIf(not utc.ENABLE_SHT31_TESTS,
                 "sht31 tests are disabled")
class SuperviseIntegrationTest(IntegrationTest,
                               utc.SuperviseIntegrationTest):
    """
    Test supervise functionality of sht31.py.
    """
    def setUp(self):
        super().setUp()
        self.setUpIntTest()


@unittest.skipIf(not utc.ENABLE_SHT31_TESTS,
                 "sht31 tests are disabled")
class PerformanceIntegrationTest(IntegrationTest,
                                 utc.PerformanceIntegrationTest):
    """
    Test performance of in honeywell.py.
    """
    def setUp(self):
        super().setUp()
        self.setUpIntTest()
        # network timing measurement
        # mean timing = 0.5 sec per measurement plus 0.75 sec overhead
        self.timeout_limit = (6.0 * 0.1 +
                              (sht31_config.MEASUREMENTS * 0.5 + 0.75))

        # temperature and humidity repeatability measurements
        # settings below are tuned short term repeatability assessment
        # assuming sht31_config.measurements = 10
        self.temp_stdev_limit = 0.5  # 1 sigma temp repeatability limit in F
        self.temp_repeatability_measurements = 30  # number of temp msmts.
        self.humidity_stdev_limit = 0.5  # 1 sigma humid repeat. limit %RH
        self.humidity_repeatability_measurements = 30  # number of temp msmts.
        self.poll_interval_sec = 1  # delay between repeatability measurements


if __name__ == "__main__":
    util.log_msg.debug = True
    unittest.main(verbosity=2)



"""
mmm config file.
"""
ALIAS = "mmm50"

# 3m50 thermostat IP addresses (on local net)
# user should configure these zones and IP addresses for their application.
MAIN_3M50 = 0  # zone 0
BASEMENT_3M50 = 1  # zone 1

# all environment variables specific to this thermostat type
env_variables = {
}

# min required env variables on all runs
required_env_variables = {
}

# supported thermostat configs
supported_configs = {"module": "mmm",
                     "type": 2,
                     "zones": [0, 1],
                     "modes": ["OFF_MODE", "HEAT_MODE", "COOL_MODE",
                               "UNKNOWN_MODE"]}

# metadata dict
# 'zone_name' is returned by self.get_zone_name
# 'host_name' is used for dns lookup of IP address for each zone
# 'ip_address' key (if present is used for hard-coding IP address
metadata = {
    MAIN_3M50: {"zone_name": "Main Level",
                "host_name": "thermostat-fd-b3-be.lan",
                },
    BASEMENT_3M50: {"zone_name": "Basement",
                    "host_name": "thermostat-27-67-11.lan",
                    }
}

default_zone = supported_configs["zones"][0]
default_zone_name = metadata[default_zone]["zone_name"]

argv = [
    "supervise.py",  # module
    ALIAS,  # thermostat
    str(default_zone),  # zone
    "18",  # poll time in sec
    "358",  # reconnect time in sec
    "2",  # tolerance
    "OFF_MODE",  # thermostat mode
    "2",  # number of measurements
    ]



"""
kumolocal config file.
"""
ALIAS = "kumolocal"

# thermostat zones
MAIN_LEVEL = 0  # zone 0
BASEMENT = 1  # zone 1

# constants
MAX_HEAT_SETPOINT = 68
MIN_COOL_SETPOINT = 70

# all environment variables specific to this thermostat type
env_variables = {
    "KUMO_USERNAME": None,
    "KUMO_PASSWORD": None,
}

# min required env variables on all runs
required_env_variables = {
    'KUMO_USERNAME': None,
    'KUMO_PASSWORD': None,
}

# supported thermostat configs
supported_configs = {"module": "kumolocal",
                     "type": 5,
                     "zones": [MAIN_LEVEL, BASEMENT],
                     "modes": ["OFF_MODE", "HEAT_MODE", "COOL_MODE",
                               "DRY_MODE", "AUTO_MODE", "UNKNOWN_MODE"]}

# metadata dict
# 'zone_name' is a placeholder, used at Thermostat class level.
# 'zone_name' is updated by device memory via Zone.get_zone_name()
# 'host_name' is used for DNS lookup to determine if device
# 'ip_address' is just for reference.
metadata = {
    MAIN_LEVEL: {"ip_address": "192.168.86.229",  # local IP, for ref only.
                 "zone_name": "Main Level",  # customize for your site.
                 "host_name": "tbd",  # used for DNS lookup
                 },
    BASEMENT: {"ip_address": "192.168.86.236",  # local IP, for ref only.
               "zone_name": "Basement",  # customize for your site.
               "host_name": "tbd",  # used for DNS lookup
               },
}

default_zone = supported_configs["zones"][0]
default_zone_name = metadata[default_zone]["zone_name"]

argv = [
    "supervise.py",  # module
    ALIAS,  # thermostat
    str(default_zone),  # zone
    "16",  # poll time in sec
    "356",  # reconnect time in sec
    "4",  # tolerance
    "OFF_MODE",  # thermostat mode
    "2",  # number of measurements
    ]



"""
Tests for thermostat_common.py
"""
# built-in imports
import operator
import pprint
import random
import unittest

# local imports
from thermostatsupervisor import thermostat_api as api
from thermostatsupervisor import thermostat_common as tc
from thermostatsupervisor import utilities as util
from tests import unit_test_common as utc


class Test(utc.UnitTest):
    """Test functions in thermostat_common.py."""

    # initialization
    switch_pos_bckup = None
    is_heat_mode_bckup = None
    is_cool_mode_bckup = None
    heat_raw_bckup = None
    schedule_heat_sp_bckup = None
    cool_raw_bckup = None
    schedule_cool_sp_bckup = None
    get_humid_support_bckup = None
    switch_position_backup = None
    revert_setpoint_func_bckup = None

    def setUp(self):
        super().setUp()
        self.setup_mock_thermostat_zone()

    def tearDown(self):
        self.teardown_mock_thermostat_zone()
        super().tearDown()

    def test_print_all_thermostat_meta_data(self):
        """
        Verify print_all_thermostat_metadata() runs without error.
        """
        self.Thermostat.print_all_thermostat_metadata(
            api.uip.get_user_inputs(api.uip.zone_name, "zone"))

    def test_set_mode(self):
        """
        Verify set_mode() runs without error.
        """
        result = self.Zone.set_mode("bogus_mode")
        self.assertFalse(result, "Zone.set_mode() should have returned False")

    def test_store_current_mode(self):
        """
        Verify store_current_mode() runs without error.
        """
        backup_func = None

        def dummy_true():
            """Return 1."""
            return 1

        test_cases = [["is_heat_mode", self.Zone.HEAT_MODE],
                      ["is_cool_mode", self.Zone.COOL_MODE],
                      ["is_dry_mode", self.Zone.DRY_MODE],
                      ["is_auto_mode", self.Zone.AUTO_MODE],
                      ["is_fan_mode", self.Zone.FAN_MODE],
                      ["is_off_mode", self.Zone.OFF_MODE]]

        print(f"thermostat_type={self.Zone.thermostat_type}")

        for test_case in test_cases:
            print(f"testing {test_case[0]}")
            try:
                # mock up the is_X_mode() functions
                if test_case[0]:
                    backup_func = getattr(self.Zone, test_case[0])
                    setattr(self.Zone, test_case[0], dummy_true)
                print(f"current mode(pre)={self.Zone.current_mode}")

                # store the current mode and check cache
                self.Zone.store_current_mode()
                print(f"current mode(post)={self.Zone.current_mode}")
                self.assertEqual(test_case[1], self.Zone.current_mode,
                                 f"Zone.store_current_mode() failed to cache"
                                 f" mode={test_case[1]}")

                # confirm verify_current_mode()
                none_act = self.Zone.verify_current_mode(None)
                self.assertTrue(none_act,
                                "verify_current_mode(None) failed "
                                "to return True")
                curr_act = self.Zone.verify_current_mode(test_case[1])
                self.assertTrue(curr_act,
                                "verify_current_mode() doesn't match "
                                "current test mode")
                dummy_act = self.Zone.verify_current_mode("dummy_mode")
                self.assertFalse(dummy_act,
                                 "verify_current_mode('dummy_mode') returned "
                                 "True, should have returned False")
            finally:
                # restore mocked function
                if test_case[0]:
                    setattr(self.Zone, test_case[0], backup_func)

    def test_check_return_types(self):
        """
        Verify return type of each function is as expected.
        """
        func_dict = {
            "is_temp_deviated_from_schedule": {
                "key": self.Zone.is_temp_deviated_from_schedule,
                "args": None,
                "return_type": bool},
            "get_current_mode": {
                "key": self.Zone.get_current_mode,
                "args": [1, 1],  # flag_all_deviations==False
                "return_type": dict},
            "Get_current_mode": {  # Capitalize for unique key
                "key": self.Zone.get_current_mode,
                "args": [1, 1, True, True],  # flag_all_deviations==True
                "return_type": dict},
            "set_mode": {
                "key": self.Zone.set_mode,
                "args": ["bogus"],
                "return_type": bool},
            "store_current_mode": {
                "key": self.Zone.store_current_mode,
                "args": None,
                "return_type": type(None)},
            "validate_numeric": {
                "key": self.Zone.validate_numeric,
                "args": [0, "bogus"],
                "return_type": int},
            "warn_if_outside_global_limit": {
                "key": self.Zone.warn_if_outside_global_limit,
                "args": [0, 0, operator.gt, "bogus"],
                "return_type": bool},
            "is_heat_mode": {
                "key": self.Zone.is_heat_mode,
                "args": None,
                "return_type": int},
            "is_cool_mode": {
                "key": self.Zone.is_cool_mode,
                "args": None,
                "return_type": int},
            "is_dry_mode": {
                "key": self.Zone.is_dry_mode,
                "args": None,
                "return_type": int},
            "is_auto_mode": {
                "key": self.Zone.is_auto_mode,
                "args": None,
                "return_type": int},
            "is_fan_mode": {
                "key": self.Zone.is_fan_mode,
                "args": None,
                "return_type": int},
            "is_off_mode": {
                "key": self.Zone.is_off_mode,
                "args": None,
                "return_type": int},
            "is_heating": {
                "key": self.Zone.is_heating,
                "args": None,
                "return_type": int},
            "is_cooling": {
                "key": self.Zone.is_cooling,
                "args": None,
                "return_type": int},
            "is_drying": {
                "key": self.Zone.is_drying,
                "args": None,
                "return_type": int},
            "is_auto": {
                "key": self.Zone.is_auto,
                "args": None,
                "return_type": int},
            "get_display_temp": {
                "key": self.Zone.get_display_temp,
                "args": None,
                "return_type": float},
            "get_display_humidity": {
                "key": self.Zone.get_display_humidity,
                "args": None,
                "return_type": float},
            "get_is_humidity_supported": {
                "key": self.Zone.get_is_humidity_supported,
                "args": None,
                "return_type": bool},
            "get_system_switch_position": {
                "key": self.Zone.get_system_switch_position,
                "args": None,
                "return_type": int},
            "get_heat_setpoint_raw": {
                "key": self.Zone.get_heat_setpoint_raw,
                "args": None,
                "return_type": int},
            "get_schedule_heat_sp": {
                "key": self.Zone.get_schedule_heat_sp,
                "args": None,
                "return_type": int},
            "get_cool_setpoint_raw": {
                "key": self.Zone.get_cool_setpoint_raw,
                "args": None,
                "return_type": int},
            "get_schedule_cool_sp": {
                "key": self.Zone.get_schedule_cool_sp,
                "args": None,
                "return_type": int},
            "get_is_invacation_hold_mode": {
                "key": self.Zone.get_is_invacation_hold_mode,
                "args": None,
                "return_type": bool},
            "get_temporary_hold_until_time": {
                "key": self.Zone.get_temporary_hold_until_time,
                "args": None,
                "return_type": int},
            "refresh_zone_info": {
                "key": self.Zone.refresh_zone_info,
                "args": None,
                "return_type": type(None)},
            "report_heating_parameters": {
                "key": self.Zone.report_heating_parameters,
                "args": None,
                "return_type": type(None)},
            "update_runtime_parameters": {
                "key": self.Zone.update_runtime_parameters,
                "args": None,
                "return_type": type(None)},
            "get_schedule_program_heat": {
                "key": self.Zone.get_schedule_program_heat,
                "args": None,
                "return_type": dict},
            "get_schedule_program_cool": {
                "key": self.Zone.get_schedule_program_cool,
                "args": None,
                "return_type": dict},
            "get_vacation_hold_until_time": {
                "key": self.Zone.get_vacation_hold_until_time,
                "args": None,
                "return_type": int},
            "set_heat_setpoint": {
                "key": self.Zone.set_heat_setpoint,
                "args": [0],
                "return_type": type(None)},
            "set_cool_setpoint": {
                "key": self.Zone.set_cool_setpoint,
                "args": [0],
                "return_type": type(None)},
            "revert_temperature_deviation": {
                "key": self.Zone.revert_temperature_deviation,
                "args": [0, "this is a dummy msg from unit test"],
                "return_type": type(None)},
        }
        for key, value in func_dict.items():
            print(f"key={key}")
            print(f"value={value}")
            expected_type = value["return_type"]
            print(f"expected type={expected_type}")
            if value["args"] is not None:
                return_val = value["key"](*value["args"])
            else:
                return_val = value["key"]()
            self.assertTrue(isinstance(return_val, expected_type),
                            f"func={key}, expected type={expected_type}, "
                            f"actual type={type(return_val)}")

    def test_validate_numeric(self):
        """Test validate_numeric() function."""
        for test_case in [1, 1.0, "1", True, None]:
            print(f"test case={type(test_case)}")
            if isinstance(test_case, (int, float)):
                expected_val = test_case
                actual_val = self.Zone.validate_numeric(test_case,
                                                        "test_case")
                self.assertEqual(
                    expected_val, actual_val,
                    f"expected return value={expected_val}, "
                    f"type({type(expected_val)}), "
                    f"actual={actual_val},type({type(actual_val)})")
            else:
                with self.assertRaises(TypeError):
                    print("attempting to input bad parameter type, "
                          "expect exception...")
                    self.Zone.validate_numeric(
                        test_case, "test_case")

    def test_warn_if_outside_global_limit(self):
        """Test warn_if_outside_global_limit() function."""
        self.assertTrue(self.Zone.warn_if_outside_global_limit(
            self.Zone.max_scheduled_heat_allowed + 1,
            self.Zone.max_scheduled_heat_allowed, operator.gt, "heat"),
            "function result should have been True")
        self.assertFalse(self.Zone.warn_if_outside_global_limit(
            self.Zone.max_scheduled_heat_allowed - 1,
            self.Zone.max_scheduled_heat_allowed, operator.gt, "heat"),
            "function result should have been False")
        self.assertTrue(self.Zone.warn_if_outside_global_limit(
            self.Zone.min_scheduled_cool_allowed - 1,
            self.Zone.min_scheduled_cool_allowed, operator.lt, "cool"),
            "function result should have been True")
        self.assertFalse(self.Zone.warn_if_outside_global_limit(
            self.Zone.min_scheduled_cool_allowed + 1,
            self.Zone.min_scheduled_cool_allowed, operator.lt, "cool"),
            "function result should have been False")

    def test_revert_thermostat_mode(self):
        """
        Test the revert_thermostat_mode() function.
        """
        test_cases = [self.Zone.HEAT_MODE, self.Zone.COOL_MODE,
                      self.Zone.DRY_MODE, self.Zone.AUTO_MODE,
                      self.Zone.FAN_MODE, self.Zone.OFF_MODE,
                      self.Zone.UNKNOWN_MODE]
        for test_case in random.choices(test_cases, k=20):
            if ((self.Zone.current_mode in self.Zone.heat_modes and
                 test_case in self.Zone.cool_modes)
                or (self.Zone.current_mode in self.Zone.cool_modes and
                    test_case in self.Zone.heat_modes)):
                expected_mode = self.Zone.OFF_MODE
            else:
                expected_mode = test_case
            print(f"reverting to '{test_case}' mode, "
                  f"expected mode={expected_mode}")
            new_mode = self.Zone.revert_thermostat_mode(test_case)
            self.assertEqual(new_mode, expected_mode,
                             f"reverting to {test_case} mode failed, new mode"
                             f" is '{new_mode}', expected '{expected_mode}'")
            self.Zone.current_mode = test_case

    def test_measure_thermostat_response_time(self):
        """
        Test the measure_thermostat_response_time() function.
        """
        # measure thermostat response time
        measurements = 3
        print(f"Thermostat response times for {measurements} measurements...")
        meas_data = self.Zone.measure_thermostat_response_time(measurements)
        ppp = pprint.PrettyPrinter(indent=4)
        ppp.pprint(meas_data)
        self.assertTrue(isinstance(meas_data, dict),
                        f"return data is type({type(meas_data)}), "
                        f"expected a dict")
        self.assertEqual(meas_data["measurements"], measurements,
                         f"number of measurements in return data("
                         f"{meas_data['measurements']}) doesn't match number "
                         f"of masurements requested({measurements})")

    def test_get_current_mode(self):
        """
        Verify get_current_mode runs in all permutations.

        test cases:
        1. heat mode and following schedule
        2. heat mode and deviation
        3. cool mode and following schedule
        4. cool mode and cool deviation
        5. humidity is available
        """
        test_cases = {
            "heat mode and following schedule": {
                "mode": self.Zone.HEAT_MODE,
                "humidity": False,
                "heat_mode": True,
                "cool_mode": False,
                "heat_deviation": False,
                "cool_deviation": False,
                "hold_mode": False,
            },
            "heat mode and following schedule and humidity": {
                "mode": self.Zone.HEAT_MODE,
                "humidity": True,
                "heat_mode": True,
                "cool_mode": False,
                "heat_deviation": False,
                "cool_deviation": False,
                "hold_mode": False,
            },
            "heat mode and deviation": {
                "mode": self.Zone.HEAT_MODE,
                "humidity": False,
                "heat_mode": True,
                "cool_mode": False,
                "heat_deviation": True,
                "cool_deviation": False,
                "hold_mode": True,
            },
            "heat mode and deviation and humidity": {
                "mode": self.Zone.HEAT_MODE,
                "humidity": True,
                "heat_mode": True,
                "cool_mode": False,
                "heat_deviation": True,
                "cool_deviation": False,
                "hold_mode": True,
            },
            "cool mode and following schedule": {
                "mode": self.Zone.COOL_MODE,
                "humidity": False,
                "heat_mode": False,
                "cool_mode": True,
                "heat_deviation": False,
                "cool_deviation": False,
                "hold_mode": False,
            },
            "cool mode and following schedule and humidity": {
                "mode": self.Zone.COOL_MODE,
                "humidity": True,
                "heat_mode": False,
                "cool_mode": True,
                "heat_deviation": False,
                "cool_deviation": False,
                "hold_mode": False,
            },
            "cool mode and deviation": {
                "mode": self.Zone.COOL_MODE,
                "humidity": False,
                "heat_mode": False,
                "cool_mode": True,
                "heat_deviation": False,
                "cool_deviation": True,
                "hold_mode": True,
            },
            "cool mode and deviation and humidity": {
                "mode": self.Zone.COOL_MODE,
                "humidity": True,
                "heat_mode": False,
                "cool_mode": True,
                "heat_deviation": False,
                "cool_deviation": True,
                "hold_mode": True,
            },
        }

        self.backup_functions()
        try:
            for test_case in test_cases:
                # mock up mode, set points, and humidity setting
                self.mock_set_mode(test_cases[test_case]["mode"])
                self.mock_set_point_deviation(
                    test_cases[test_case]["heat_deviation"],
                    test_cases[test_case]["cool_deviation"])
                self.mock_set_humidity_support(
                    test_cases[test_case]["humidity"])

                # call function and print return value
                ret_dict = self.Zone.get_current_mode(1, 1, True, False)
                print(f"test case '{test_case}' result: '{ret_dict}'")

                # verify return states are correct
                for return_val in ["heat_mode", "cool_mode", "heat_deviation",
                                   "cool_deviation", "hold_mode"]:
                    self.assertEqual(ret_dict[return_val],
                                     test_cases[test_case][return_val],
                                     f"test case '{test_case}' parameter "
                                     f"'{return_val}', result="
                                     f"{ret_dict[return_val]}, expected="
                                     f"{test_cases[test_case][return_val]}")

                # verify humidity reporting
                if test_cases[test_case]["humidity"]:
                    self.assertTrue("humidity" in ret_dict["status_msg"])
                else:
                    self.assertTrue("humidity" not in ret_dict["status_msg"])
        finally:
            self.restore_functions()

    def mock_set_mode(self, mock_mode):
        """
        Mock heat setting by overriding switch position function.

        Make sure to backup and restore methods if using this function.
        inputs:
            mock_mode(str): mode string
        returns:
            None
        """
        if mock_mode == self.Zone.HEAT_MODE:
            self.Zone.is_heat_mode = \
                (lambda *_, **__: True)
            self.Zone.is_cool_mode = \
                (lambda *_, **__: False)
            self.Zone.is_dry_mode = \
                (lambda *_, **__: False)
            self.Zone.is_auto_mode = \
                (lambda *_, **__: False)
            self.Zone.is_fan_mode = \
                (lambda *_, **__: False)
            self.Zone.is_off_mode = \
                (lambda *_, **__: False)
            self.Zone.current_mode = self.Zone.HEAT_MODE
        elif mock_mode == self.Zone.COOL_MODE:
            self.Zone.is_heat_mode = \
                (lambda *_, **__: False)
            self.Zone.is_cool_mode = \
                (lambda *_, **__: True)
            self.Zone.is_dry_mode = \
                (lambda *_, **__: False)
            self.Zone.is_auto_mode = \
                (lambda *_, **__: False)
            self.Zone.is_fan_mode = \
                (lambda *_, **__: False)
            self.Zone.is_off_mode = \
                (lambda *_, **__: False)
            self.Zone.current_mode = self.Zone.COOL_MODE
        elif mock_mode == self.Zone.DRY_MODE:
            self.Zone.is_heat_mode = \
                (lambda *_, **__: False)
            self.Zone.is_cool_mode = \
                (lambda *_, **__: False)
            self.Zone.is_dry_mode = \
                (lambda *_, **__: True)
            self.Zone.is_auto_mode = \
                (lambda *_, **__: False)
            self.Zone.is_fan_mode = \
                (lambda *_, **__: False)
            self.Zone.is_off_mode = \
                (lambda *_, **__: False)
            self.Zone.current_mode = self.Zone.DRY_MODE
        elif mock_mode == self.Zone.AUTO_MODE:
            self.Zone.is_heat_mode = \
                (lambda *_, **__: False)
            self.Zone.is_cool_mode = \
                (lambda *_, **__: False)
            self.Zone.is_dry_mode = \
                (lambda *_, **__: False)
            self.Zone.is_auto_mode = \
                (lambda *_, **__: True)
            self.Zone.is_fan_mode = \
                (lambda *_, **__: False)
            self.Zone.is_off_mode = \
                (lambda *_, **__: False)
            self.Zone.current_mode = self.Zone.AUTO_MODE
        elif mock_mode == self.Zone.FAN_MODE:
            self.Zone.is_heat_mode = \
                (lambda *_, **__: False)
            self.Zone.is_cool_mode = \
                (lambda *_, **__: False)
            self.Zone.is_dry_mode = \
                (lambda *_, **__: False)
            self.Zone.is_auto_mode = \
                (lambda *_, **__: False)
            self.Zone.is_fan_mode = \
                (lambda *_, **__: True)
            self.Zone.is_off_mode = \
                (lambda *_, **__: False)
            self.Zone.current_mode = self.Zone.FAN_MODE
        elif mock_mode == self.Zone.OFF_MODE:
            self.Zone.is_heat_mode = \
                (lambda *_, **__: False)
            self.Zone.is_cool_mode = \
                (lambda *_, **__: False)
            self.Zone.is_dry_mode = \
                (lambda *_, **__: False)
            self.Zone.is_auto_mode = \
                (lambda *_, **__: False)
            self.Zone.is_fan_mode = \
                (lambda *_, **__: False)
            self.Zone.is_off_mode = \
                (lambda *_, **__: True)
            self.Zone.current_mode = self.Zone.OFF_MODE
        else:
            self.fail(f"mock mode '{mock_mode}' is not supported")

    def mock_set_point_deviation(self, heat_deviation, cool_deviation):
        """
        Override heat and cool set points with mock values.

        inputs:
            heat_deviation(bool): True if heat is deviated
            cool_deviation(bool): True if cool is deviated
        returns:
            None
        """
        deviation_val = self.Zone.tolerance_degrees + 1
        heat_sched_sp = self.Zone.max_scheduled_heat_allowed - 13
        heat_sp = heat_sched_sp + [0, deviation_val][heat_deviation]
        cool_sched_sp = self.Zone.min_scheduled_cool_allowed + 13
        cool_sp = cool_sched_sp - [0, deviation_val][cool_deviation]

        self.Zone.get_heat_setpoint_raw = (lambda *_, **__: heat_sp)
        self.Zone.get_schedule_heat_sp = (lambda *_, **__: heat_sched_sp)
        self.Zone.get_cool_setpoint_raw = (lambda *_, **__: cool_sp)
        self.Zone.get_schedule_cool_sp = (lambda *_, **__: cool_sched_sp)

    def mock_set_humidity_support(self, bool_val):
        """
        Mock humidity support.

        inputs:
            bool_val(bool): humidity support state
        returns:
            None
        """
        self.Zone.get_is_humidity_supported = \
            (lambda *_, **__: bool_val)

    def backup_functions(self):
        """Backup functions prior to mocking return values."""
        self.switch_pos_bckup = self.Zone.get_system_switch_position
        self.is_heat_mode_bckup = self.Zone.is_heat_mode
        self.is_cool_mode_bckup = self.Zone.is_cool_mode
        self.heat_raw_bckup = self.Zone.get_heat_setpoint_raw
        self.schedule_heat_sp_bckup = self.Zone.get_schedule_heat_sp
        self.cool_raw_bckup = self.Zone.get_cool_setpoint_raw
        self.schedule_cool_sp_bckup = self.Zone.get_schedule_cool_sp
        self.get_humid_support_bckup = self.Zone.get_is_humidity_supported
        self.revert_setpoint_func_bckup = self.Zone.revert_setpoint_func

    def restore_functions(self):
        """Restore backed up functions."""
        self.Zone.get_system_switch_position = self.switch_pos_bckup
        self.Zone.is_heat_mode = self.is_heat_mode_bckup
        self.Zone.is_cool_mode = self.is_cool_mode_bckup
        self.Zone.get_heat_setpoint_raw = self.heat_raw_bckup
        self.Zone.get_schedule_heat_sp = self.schedule_heat_sp_bckup
        self.Zone.get_cool_setpoint_raw = self.cool_raw_bckup
        self.Zone.get_schedule_cool_sp = self.schedule_cool_sp_bckup
        self.Zone.get_is_humidity_supported = self.get_humid_support_bckup
        self.Zone.revert_setpoint_func = self.revert_setpoint_func_bckup

    def test_display_basic_thermostat_summary(self):
        """Confirm print_basic_thermostat_summary() works without error."""

        # override switch position function to be determinant
        self.switch_position_backup = self.Zone.get_system_switch_position
        try:
            self.Zone.get_system_switch_position = \
                (lambda *_, **__: self.Zone.system_switch_position[
                    tc.ThermostatCommonZone.DRY_MODE])
            self.Zone.display_basic_thermostat_summary()
        finally:
            self.Zone.get_system_switch_position = self.switch_position_backup

    @unittest.skipIf(not utc.ENABLE_SHT31_TESTS,
                     "sht31 tests are disabled")
    def test_thermostat_basic_checkout(self):
        """Verify thermostat_basic_checkout()."""

        # override switch position function to be determinant
        self.switch_position_backup = self.Zone.get_system_switch_position
        try:
            self.Zone.get_system_switch_position = \
                (lambda *_, **__: self.Zone.system_switch_position[
                    tc.ThermostatCommonZone.DRY_MODE])
            api.uip = api.UserInputs(self.unit_test_argv)
            thermostat_type = api.uip.get_user_inputs(
                api.uip.zone_name, api.input_flds.thermostat_type)
            zone_number = api.uip.get_user_inputs(api.uip.zone_name,
                                                  api.input_flds.zone)
            mod = api.load_hardware_library(thermostat_type)
            thermostat, zone = \
                tc.thermostat_basic_checkout(
                    thermostat_type,
                    zone_number,
                    mod.ThermostatClass, mod.ThermostatZone
                )
            print(f"thermotat={type(thermostat)}")
            print(f"thermotat={type(zone)}")
        finally:
            self.Zone.get_system_switch_position = self.switch_position_backup

    def test_revert_temperature_deviation(self):
        """Verify revert_temperature_deviation()."""

        def mock_revert_setpoint_func(setpoint):
            self.Zone.current_setpoint = setpoint

        # backup functions that may be mocked
        self.backup_functions()
        try:
            # mock up the revert function for unit testing
            self.Zone.revert_setpoint_func = mock_revert_setpoint_func

            # mock the thermostat into heat mode
            # self.mock_set_mode(self.Zone.HEAT_MODE)
            # print(f"current thermostat mode={self.Zone.current_mode}")

            for new_setpoint in [13, 26, -4, 101]:
                # get current setpoint
                current_setpoint = self.Zone.current_setpoint

                # revert setpoint
                msg = (f"reverting setpoint from "
                       f"{util.temp_value_with_units(current_setpoint)} to "
                       f"{util.temp_value_with_units(new_setpoint)}")
                self.Zone.revert_temperature_deviation(new_setpoint, msg)

                # verify setpoint
                actual_setpoint = self.Zone.current_setpoint
                self.assertEqual(
                    new_setpoint, actual_setpoint,
                    f"reverting setpoint failed, actual="
                    f"{util.temp_value_with_units(actual_setpoint)}, expected="
                    f"{util.temp_value_with_units(new_setpoint)}")

            # verify function default behavior
            new_setpoint = self.Zone.current_setpoint = 56

            # revert setpoint
            msg = (f"reverting setpoint from "
                   f"{util.temp_value_with_units(actual_setpoint)} to "
                   f"{util.temp_value_with_units(new_setpoint)}")
            self.Zone.revert_temperature_deviation(msg=msg)

            # verify setpoint
            actual_setpoint = self.Zone.current_setpoint
            self.assertEqual(
                new_setpoint, actual_setpoint,
                f"reverting setpoint failed, actual="
                f"{util.temp_value_with_units(actual_setpoint)}, expected="
                f"{util.temp_value_with_units(new_setpoint)}")

        finally:
            self.restore_functions()

    def test_report_heating_parameters(self):
        """Verify report_heating_parameters()."""
        test_cases = [tc.ThermostatCommonZone.UNKNOWN_MODE,
                      tc.ThermostatCommonZone.OFF_MODE]
        for test_case in test_cases:
            print(f"test_case={test_case}")
            self.Zone.report_heating_parameters(
                switch_position=self.Zone.system_switch_position[test_case])

    def test_display_runtime_settings(self):
        """Verify display_runtime_settings()."""
        self.Zone.display_runtime_settings()

    def test_display_session_settings(self):
        """
        Verify display_session_settings() with all permutations.
        """
        for self.Zone.revert_deviations in [False, True]:
            for self.Zone.revert_all_deviations in [False, True]:
                print(f"{'-' * 60}")
                print(f"testing revert={self.Zone.revert_deviations}, "
                      f"revert all={self.Zone.revert_all_deviations}")
                self.Zone.display_session_settings()


if __name__ == "__main__":
    util.log_msg.debug = True
    unittest.main(verbosity=2)



"""Common utility functions and globals."""

# built-in libraries
import argparse
import configparser
import datetime
import inspect
import os
import platform
import socket
import sys


PACKAGE_NAME = 'thermostatsupervisor'  # should match name in __init__.py

# error codes
NO_ERROR = 0
CONNECTION_ERROR = 1
AUTHORIZATION_ERROR = 2
EMAIL_SEND_ERROR = 3
ENVIRONMENT_ERROR = 4
FILE_NOT_FOUND_ERROR = 5
OTHER_ERROR = 99

# bogus values to identify uninitialized data
BOGUS_INT = -123456789
BOGUS_BOOL = False
BOGUS_STR = "<missing value>"
bogus_dict = {}

# logging options
CONSOLE_LOG = 0x001  # print to console
DATA_LOG = 0x010  # print to data log
BOTH_LOG = 0x011  # log to both console and data logs
DEBUG_LOG = 0x100  # print only if debug mode is on

FILE_PATH = ".//data"
MAX_LOG_SIZE_BYTES = 2**20  # logs rotate at this max size
HTTP_TIMEOUT = 60  # timeout in seconds

# set unit test IP address, same as client
unit_test_mode = False  # in unit test mode


def get_function_name(stack_value=1):
    """
    Return function name from stack.

    inputs:
        stack_val(int): position in stack, 1=caller, 2=caller's parent
    returns:
        (str): function name
    """
    return inspect.stack()[stack_value][3]


def log_msg(msg, mode, func_name=-1, file_name=None):
    """
    Log message to file, console, etc.

    inputs:
        msg(str): message to log
        mode(int): log destination(s), see logging options at top of file
        func_name(int): if > 1, will include func name from this stack position
        file_name(str): if None will use default name
    returns:
        (dict): {status, tbd}
    """
    return_buffer = {
        "status": NO_ERROR,
    }

    # set debug mode
    debug_enabled = getattr(log_msg, "debug", False)
    debug_msg = mode & DEBUG_LOG
    filter_debug_msg = debug_msg and not debug_enabled

    # define filename
    if file_name is not None:
        log_msg.file_name = file_name

    # build message string
    if func_name > 0:
        msg = f"[{get_function_name(func_name)}]: {msg}"

    # log to data file
    if (mode & DATA_LOG) and not filter_debug_msg:
        # create directory if needed
        if not os.path.exists(FILE_PATH):
            print(f"data folder '{FILE_PATH}' created.")
            os.makedirs(FILE_PATH)

        # build full file name
        full_path = get_full_file_path(log_msg.file_name)

        # check file size and rotate if necessary
        file_size_bytes = get_file_size_bytes(full_path)
        file_size_bytes = log_rotate_file(full_path, file_size_bytes,
                                          MAX_LOG_SIZE_BYTES)

        # write to file
        write_to_file(full_path, file_size_bytes, msg)

    # print to console
    if (mode & CONSOLE_LOG) and not filter_debug_msg:
        print(msg)

    return return_buffer


# global default log file name if none is specified
log_msg.file_name = "default_log.txt"


def get_file_size_bytes(full_path):
    """
    Get the file size for the specified log file.

    inputs:
        full_path(str): full file name and path.
    returns:
        file_size_bytes(int): file size in bytes
    """
    try:
        file_size_bytes = os.path.getsize(full_path)
    except FileNotFoundError:
        # file does not exist
        file_size_bytes = 0
    return file_size_bytes


def log_rotate_file(full_path, file_size_bytes, max_size_bytes):
    """
    Rotate log file to prevent file from getting too large.

    inputs:
        full_path(str): full file name and path.
        file_size_bytes(int): file size in bytes
        max_size_bytes(int): max allowable file size.
    returns:
        file_size_bytes(int): file size in bytes
    """
    if file_size_bytes > max_size_bytes:
        # rotate log file
        current_date = datetime.datetime.today().strftime(
            '%d-%b-%Y-%H-%M-%S')
        os.rename(full_path, full_path[:-4] + "-" +
                  str(current_date) + '.txt')
        file_size_bytes = 0
    return file_size_bytes


def write_to_file(full_path, file_size_bytes, msg):
    """
    Rotate log file to prevent file from getting too large.

    inputs:
        full_path(str): full file name and path.
        file_size_bytes(int): file size in bytes
        msg(str): message to write.
    returns:
        (int): number of bytes written to file.
    """
    if file_size_bytes == 0:
        write_mode = "w"  # writing
    else:
        write_mode = "a"  # appending
    with open(full_path, write_mode, encoding="utf8") as file_handle:
        msg_to_write = msg + "\n"
        file_handle.write(msg_to_write)
    return utf8len(msg_to_write)


def is_windows_environment():
    """Return True if running on Windows PC."""
    print(f"DEBUG: platform={platform.system().upper()}")
    return 'WINDOWS' in platform.system().upper()


def get_full_file_path(file_name):
    """
    Return full file path.

    inputs:
        file_name(str): name of file with extension
    returns:
        (str) full file name and path
    """
    return FILE_PATH + "//" + file_name


def utf8len(input_string):
    """
    Return length of string in bytes.

    inputs:
        input_string(str): input string.
    returns:
        (int): length of string in bytes.
    """
    return len(input_string.encode('utf-8'))


def temp_value_with_units(raw, disp_unit='F', precision=1) -> str:
    """
    Return string representing temperature and units.

    inputs:
        raw(int or float): temperature value.
        disp_unit(str): display unit character.
        precision(int): number of digits after decimal.
    returns:
        (str): temperature and units.
    """
    if disp_unit.upper() not in ['C', 'F', 'K']:
        raise ValueError(f"{get_function_name()}: '{disp_unit}' is not a "
                         "valid temperature unit")

    # if string try to convert to float
    if isinstance(raw, str):
        if '°' in raw:
            return raw  # pass-thru
        try:
            raw = float(raw)
        except ValueError:
            pass

    if raw is None:
        formatted = f"{raw}"
    elif precision == 0:
        formatted = f"{raw:.0f}"
    else:
        formatted = f"{raw:.{precision}f}"
    return f'{formatted}°{disp_unit}'


def humidity_value_with_units(raw, disp_unit=' RH', precision=0) -> str:
    """
    Return string representing humidity and units.

    inputs:
        raw(int or float): humidity value.
        disp_unit(str): display unit character.
        precision(int): number of digits after decimal.
    returns:
        (str): temperature and units.
    """
    if disp_unit.upper() not in ['RH', ' RH']:
        raise ValueError(f"{get_function_name()}: '{disp_unit}' is not a "
                         " valid humidity unit")

    # if string try to convert to float
    if isinstance(raw, str):
        if '%' in raw:
            return raw  # pass-thru
        try:
            raw = float(raw)
        except ValueError:
            pass

    if raw is None:
        formatted = f"{raw}"
    elif precision == 0:
        formatted = f"{raw:.0f}"
    else:
        formatted = f"{raw:.{precision}f}"
    return f'{formatted}%{disp_unit}'


def get_key_from_value(input_dict, val):
    """
    Return first key found in dict from value provided.

    inputs:
        input_dict(dict): target dictionary
        val(str or int):  value
    returns:
        (str or int): dictionary key
    """
    for key, value in input_dict.items():
        if val == value:
            return key
    raise KeyError(f"key not found in dict '{input_dict}' with value='{val}'")


def c_to_f(tempc) -> float:
    """
    Convert from Celsius to Fahrenheit.

    inputs:
        tempc(int, float): temp in deg c.
    returns:
        (float): temp in deg f.
    """
    if isinstance(tempc, type(None)):
        return tempc  # pass thru
    elif isinstance(tempc, (int, float)):
        return tempc * 9.0 / 5 + 32
    else:
        raise TypeError(f"raw value '{tempc}' is not an int or float")


def f_to_c(tempf) -> float:
    """
    Convert from Fahrenheit to Celsius.

    inputs:
        tempc(int, float): temp in deg f.
    returns:
        (float): temp in deg c.
    """
    if isinstance(tempf, type(None)):
        return tempf  # pass thru
    elif isinstance(tempf, (int, float)):
        return (tempf - 32) * 5 / 9.0
    else:
        raise TypeError(f"raw value '{tempf}' is not an int or float")


def is_host_on_local_net(host_name, ip_address=None, verbose=False):
    """
    Return True if specified host is on local network.
    socket.gethostbyaddr() throws exception for some IP address
    so preferred way to use this function is to pass in only the
    hostname and leave the IP as default (None).
    inputs:
        host_name(str): expected host name.
        ip_address(str): target IP address on local net.
        verbose(bool): if True, print out status.
    returns:
        tuple(bool, str): True if confirmed on local net, else False.
                          ip_address if known
    """
    host_found = None
    # find by hostname alone if IP is None
    if ip_address is None:
        try:
            host_found = socket.gethostbyname(host_name)
        except socket.gaierror:
            return False, None
        if host_found:
            if verbose:
                print(f"host {host_name} found at {host_found} on local net")
            return True, host_found
        else:
            if verbose:
                print(f"host {host_name} is not detected on local net")
            return False, None

    else:
        # match both IP and host if both are provided.
        try:
            host_found = socket.gethostbyaddr(ip_address)
        except socket.herror:  # exception if DNS name is not set
            return False, None
        if host_name == host_found[0]:
            return True, ip_address
        else:
            print(f"DEBUG: expected host={host_name}, "
                  f"actual host={host_found}")
            return False, None


# default parent_key if user_inputs are not pulled from file
default_parent_key = "argv"


class UserInputs():
    """Manage runtime arguments."""

    def __init__(self, argv_list, help_description, suppress_warnings=False,
                 parent_key=default_parent_key,
                 *_, **__):
        """
        Base Class UserInputs Constructor.

        user_inputs is a dictionary of runtime parameters.
        structure = {<parent_key> : {<child_key: {}}
        dict can have multiple parent_keys and multiple child_keys

        inputs:
            argv_list(list): override runtime values.
            help_description(str): description field for help text.
            suppress_warnings(bool): True to suppress warning msgs.
            parent_key(str, int): parent key
        """
        self.argv_list = argv_list
        self.default_parent_key = parent_key
        self.parent_keys = [parent_key]
        self.help_description = help_description
        self.suppress_warnings = suppress_warnings
        self.parser = argparse.ArgumentParser(
            description=self.help_description)
        self.user_inputs = {}
        self.user_inputs_file = {}
        self.using_input_file = False
        self.initialize_user_inputs()
        # parse the runtime arguments from input list or sys.argv
        self.parse_runtime_parameters(argv_list)

    def initialize_user_inputs(self, parent_keys=None):
        """Populate user_inputs dictionary."""
        pass  # placeholder, is instance-specific

    def get_sflag_list(self):
        """Return a list of all sflags."""
        valid_sflags = []
        for parent_key, child_dict in self.user_inputs.items():
            for child_key, _ in child_dict.items():
                valid_sflags.append(self.user_inputs[parent_key][
                    child_key]["sflag"])
        return valid_sflags

    def parse_runtime_parameters(self, argv_list=None):
        """
        Parse all runtime parameters from list, argv list or named
        arguments.

        If argv_list is input then algo will default to input list.
        ElIf hyphen is found in argv the algo will default to named args.
        inputs:
           argv_list: list override for sys.argv
        returns:
          argv_dict(dict)
        """
        sysargv_sflags = [str(elem)[:2] for elem in sys.argv[1:]]
        if self.user_inputs is None:
            raise ValueError("user_inputs cannot be None")
        parent_key = list(self.user_inputs.keys())[0]
        valid_sflags = self.get_sflag_list()
        valid_sflags += ["-h", "--"]  # add help and double dash
        if argv_list:
            # argument list input, support parsing list
            argvlist_sflags = [str(elem)[:2] for elem in argv_list]
            if any([flag in argvlist_sflags for flag in valid_sflags]):
                log_msg(
                    f"parsing named runtime parameters from user input list: "
                    f"{argv_list}",
                    mode=DEBUG_LOG +
                    CONSOLE_LOG,
                    func_name=1)
                self.parse_named_arguments(argv_list=argv_list)
            else:
                log_msg(
                    f"parsing runtime parameters from user input list: "
                    f"{argv_list}",
                    mode=DEBUG_LOG +
                    CONSOLE_LOG,
                    func_name=1)
                self.parse_argv_list(
                    parent_key, argv_list)
        elif any([flag in sysargv_sflags for flag in valid_sflags]):
            # named arguments from sys.argv
            log_msg(
                f"parsing named runtime parameters from sys.argv: {sys.argv}",
                mode=DEBUG_LOG +
                CONSOLE_LOG,
                func_name=1)
            self.parse_named_arguments()
        else:
            # sys.argv parsing
            log_msg(
                f"parsing runtime parameters from sys.argv: {sys.argv}",
                mode=DEBUG_LOG +
                CONSOLE_LOG,
                func_name=1)
            self.parse_argv_list(parent_key, sys.argv)

        # dynamically update valid range and defaults
        # also can trigger input file parsing based on input flags
        self.dynamic_update_user_inputs()

        # validate inputs
        self.validate_argv_inputs(self.user_inputs)

        return self.user_inputs

    def parse_named_arguments(self, parent_key=None, argv_list=None):
        """
        Parse all possible named arguments.

        inputs:
            parent_key(str): parent key for dict.
            argv_list(list): override sys.argv (for testing)
        returns:
            (dict) of all runtime parameters.
        """
        # set parent key
        if parent_key is None:
            parent_key = self.default_parent_key

        # load parser contents
        for _, attr in self.user_inputs[parent_key].items():
            self.parser.add_argument(attr["lflag"], attr["sflag"],
                                     default=attr["default"],
                                     type=attr["type"],
                                     required=attr["required"],
                                     help=attr["help"]
                                     )
        # parse the argument list
        if argv_list is not None:
            # test mode, override sys.argv
            args = self.parser.parse_args(argv_list[1:])
        else:
            args = self.parser.parse_args()
        for key in self.user_inputs[parent_key]:
            if key == "script":
                # add script name
                self.user_inputs[parent_key][key]["value"] = sys.argv[0]
            else:
                self.user_inputs[parent_key][key]["value"] = getattr(args,
                                                                     key, None)
                strip_types = str
                if isinstance(self.user_inputs[parent_key][key]["value"],
                              strip_types):
                    # str parsing has leading spaces for some reason
                    self.user_inputs[parent_key][key]["value"] = \
                        self.user_inputs[parent_key][key]["value"].strip()

        return self.user_inputs

    def parse_argv_list(self, parent_key, argv_list=None):
        """
        Parse un-named arguments from list.

        inputs:
            parent_key(str): parent key in the user_inputs dict.
            argv_list(list): list of runtime arguments in the order
                             argv_list[0] should be script name.
                             speci5fied in argv_dict "order" fields.
        returns:
            (dict) of all runtime parameters.
        """
        # set parent key
        if parent_key is None:
            parent_key = self.default_parent_key

        # if argv list is set use that, else use sys.argv
        if argv_list:
            argv_inputs = argv_list
        else:
            argv_inputs = sys.argv

        # populate dict with values from list
        for child_key, val in self.user_inputs[parent_key].items():
            if val["order"] <= len(argv_inputs) - 1:
                if (self.user_inputs[parent_key][child_key]["type"] in
                        [int, float, str]):
                    # cast data type when reading value
                    self.user_inputs[parent_key][child_key]["value"] = (
                        self.user_inputs[parent_key][child_key][
                            "type"](argv_inputs[val["order"]]))
                else:
                    # no casting, just read raw from list
                    self.user_inputs[parent_key][child_key]["value"] = \
                        argv_inputs[val["order"]]

        return self.user_inputs

    def dynamic_update_user_inputs(self):
        """Update user_inputs dict dynamically based on runtime parameters."""
        pass  # placeholder

    def validate_argv_inputs(self, argv_dict):
        """
        Validate argv inputs and update reset to defaults if necessary.

        inputs:
            argv_dict(dict): dictionary of runtime args with these elements:
            <parent_key>: {
            <key>: {  # key = argument name
                "order": 0,  # order in the argv list
                "value": None,   # initialized to None
                "type": str,  # datatype
                "default": "supervise.py",  # default value
                "sflag": "-s",  # short flag identifier
                "lflag": "--script",  # long flag identifier
                "help": "script name"},  # help text
                "valid": None,  # valid choices
                "required": True,  # is parameter required?
            }}
        returns:
            (dict) of all runtime parameters, only needed for testing.
        """
        for parent_key, child_dict in argv_dict.items():
            for child_key, attr in child_dict.items():
                proposed_value = attr["value"]
                default_value = attr["default"]
                proposed_type = type(proposed_value)
                expected_type = attr["type"]
                # missing value check
                if proposed_value is None:
                    if not self.suppress_warnings:
                        log_msg(
                            f"parent_key={parent_key}, child_key='{child_key}'"
                            f": argv parameter missing, using default "
                            f"value '{default_value}'",
                            mode=DEBUG_LOG +
                            CONSOLE_LOG,
                            func_name=1)
                    attr["value"] = attr["default"]

                # wrong datatype check
                elif proposed_type != expected_type:
                    if not self.suppress_warnings:
                        log_msg(
                            f"parent_key={parent_key}, child_key='{child_key}'"
                            f": datatype error, expected="
                            f"{expected_type}, actual={proposed_type}, "
                            "using default value "
                            f"'{default_value}'",
                            mode=DEBUG_LOG +
                            CONSOLE_LOG,
                            func_name=1)
                    attr["value"] = attr["default"]

                # out of range check
                elif (attr["valid_range"] is not None and
                      proposed_value not in attr["valid_range"]):
                    if not self.suppress_warnings:
                        log_msg(
                            f"WARNING: '{proposed_value}' is not a valid "
                            f"choice parent_key='{parent_key}', child_key="
                            f"'{child_key}', using default '{default_value}'",
                            mode=BOTH_LOG,
                            func_name=1)
                    attr["value"] = attr["default"]

        return argv_dict

    def get_user_inputs(self, parent_key, child_key, field="value"):
        """
        Return the target key's value from user_inputs.

        inputs:
            parent_key(str): top-level key
            child_key(str): second-level key
            field(str): field name, default = "value"
        returns:
            None
        """
        if child_key is None:
            return self.user_inputs[parent_key][field]
        else:
            try:
                return self.user_inputs[parent_key][child_key][field]
            except TypeError:
                print(f"TypeError: parent_key({type(parent_key)})={parent_key}"
                      f", child_key({type(child_key)})={child_key}, "
                      f"field({type(field)})={field})")
                raise
            except KeyError:
                print(f"KeyError: target=[{parent_key}][{child_key}][{field}],"
                      f" raw={self.user_inputs.keys()}")
                raise

    def set_user_inputs(self, parent_key, child_key, input_val, field="value"):
        """
        Set the target key's value from user_inputs.

        inputs:
            parent_key(str): top-level key
            child_key(str): second-level key
            input_val(str, int, float, etc.):  value to set.
            field(str): field name, default = "value"
        returns:
            None, updates uip.user_inputs dict.
        """
        if child_key is None:
            self.user_inputs[parent_key][field] = input_val
        else:
            try:
                self.user_inputs[parent_key][child_key][field] = input_val
            except TypeError:
                print(f"TypeError: keys={self.user_inputs.keys()} "
                      f"(type={type(self.user_inputs.keys())})")
                raise
            except KeyError:
                print(f"KeyError: target=[{parent_key}][{child_key}][{field}],"
                      f" raw={self.user_inputs.keys()}")
                raise

    def is_valid_file(self, arg):
        """
        Verify file input is valid.

        inputs:
            arg(str): file name with path.
        returns:
            open file handle
        """
        arg = arg.strip()  # remove any leading spaces
        if not os.path.exists(arg):
            self.parser.error(f"The file {os.path.abspath(arg)} "
                              "does not exist!")
        else:
            return open(arg, 'r', encoding="utf8")  # return a file handle

    def parse_input_file(self, input_file):
        """
        Parse an input file into a dict.

        Primary key is the section.
        Secondary key is the parameter.
        """
        input_file = input_file.strip()  # strip any whitespace
        config = configparser.ConfigParser()
        result = config.read(os.path.join(os.getcwd(), input_file))
        if not result:
            raise FileNotFoundError(f"file '{input_file}' was not found")
        sections = config.sections()
        if not sections:
            raise ValueError("INI file must have sections")
        for section in sections:
            self.user_inputs_file[section] = {}
            for key in config[section]:
                self.user_inputs_file[section][key] = config[section][key]



"""
kumocloud config file.
"""
ALIAS = "kumocloud"

# thermostat zones
MAIN_LEVEL = 0  # zone 0
BASEMENT = 1  # zone 1

# constants
MAX_HEAT_SETPOINT = 68
MIN_COOL_SETPOINT = 70

# all environment variables specific to this thermostat type
env_variables = {
    "KUMO_USERNAME": None,
    "KUMO_PASSWORD": None,
}

# min required env variables on all runs
required_env_variables = {
    'KUMO_USERNAME': None,
    'KUMO_PASSWORD': None,
}

# supported thermostat configs
supported_configs = {"module": "kumocloud",
                     "type": 4,
                     "zones": [MAIN_LEVEL, BASEMENT],
                     "modes": ["OFF_MODE", "HEAT_MODE", "COOL_MODE",
                               "DRY_MODE", "AUTO_MODE", "UNKNOWN_MODE"]}

# metadata dict
# 'zone_name' is updated by Zone.get_zone_name()
# 'host_name' is just for reference
metadata = {
    MAIN_LEVEL: {"zone_name": "Main Level",
                 "host_name": "tbd",
                 "serial_number": None,
                 },
    BASEMENT: {"zone_name": "basement",
               "host_name": "tbd",
               "serial_number": None,
               },
}

default_zone = supported_configs["zones"][0]
default_zone_name = metadata[default_zone]["zone_name"]

argv = [
    "supervise.py",  # module
    ALIAS,  # thermostat
    str(default_zone),  # zone
    "18",  # poll time in sec
    "358",  # reconnect time in sec
    "2",  # tolerance
    "OFF_MODE",  # thermostat mode
    "2",  # number of measurements
    ]



""""
flask API for raspberry pi
example from http://www.pibits.net/code/raspberry-pi-sht31-sensor-example.php
"""

# raspberry pi libraries
try:
    from RPi import GPIO  # noqa F405 raspberry pi GPIO library
    import smbus  # noqa F405
    pi_library_exception = None  # successful
except ImportError as ex:
    # hardware-related library, not needed in unittest mode
    print("WARNING: RPi or smbus library import error, "
          "this is expected in unittest mode")
    pi_library_exception = ex  # unsuccessful

# built-in imports
import munch
import re
import statistics
import subprocess
import sys
import time

# third party imports
from flask import Flask, request
from flask_restful import Resource, Api  # noqa F405
from flask_wtf.csrf import CSRFProtect
from str2bool import str2bool

# local imports
from thermostatsupervisor import environment as env
from thermostatsupervisor import flask_generic as flg
from thermostatsupervisor import sht31_config
from thermostatsupervisor import utilities as util

# runtime override fields
input_flds = munch.Munch()
input_flds.debug_fld = "debug"

uip = None  # user inputs object

# SHT31D write commands (register, [data])
# spec: https://cdn-shop.adafruit.com/product-files/
# 2857/Sensirion_Humidity_SHT3x_Datasheet_digital-767294.pdf

# single shot mode: clock_stretching, repeatability
cs_enabled_high = (0x2c, [0x06])
cs_enabled_med = (0x2c, [0x0d])
cs_enabled_low = (0x2c, [0x10])
cs_disabled_high = (0x24, [0x00])
cs_disabled_med = (0x24, [0x0b])
cs_disabled_low = (0x24, [0x16])

# periodic data acquisition modes: mps, repeatability
mps_0p5_high = (0x20, [0x32])
mps_0p5_med = (0x20, [0x24])
mps_0p5_low = (0x20, [0x2f])
mps_1_high = (0x21, [0x30])
mps_1_med = (0x21, [0x26])
mps_1_low = (0x21, [0x2d])
mps_2_high = (0x22, [0x36])
mps_2_med = (0x22, [0x20])
mps_2_low = (0x22, [0x2b])
mps_4_high = (0x23, [0x34])
mps_4_med = (0x23, [0x22])
mps_4_low = (0x23, [0x29])
mps_10_high = (0x27, [0x37])
mps_10_med = (0x27, [0x21])
mps_10_low = (0x27, [0x2a])

# misc commands
fetch_data = (0xe0, [0x00])
period_meas_with_art = (0x2b, [0x32])
break_periodic_data = (0x30, [0x93])
soft_reset = (0x30, [0xa2])
reset = (0x00, [0x06])
enable_heater = (0x30, [0x6d])
disable_heater = (0x30, [0x66])
clear_status_register = (0x30, [0x41])
read_status_register = (0xf3, [0x2d])


class Sensors:
    """Sensor data."""

    def __init__(self):
        pass

    def convert_data(self, data):
        """
        Convert data from bits to real units.

        inputs:
            data(class 'list'): raw data structure
        returns:
            temp(int): raw temp data in bits.
            temp_c(float): temp on deg C
            temp_f(float): temp in deg F
            humidity(float): humidity in %RH
        """
        # convert the data
        temp = data[0] * 256 + data[1]
        temp_c = -45 + (175 * temp / 65535.0)
        temp_f = -49 + (315 * temp / 65535.0)
        humidity = 100 * (data[3] * 256 + data[4]) / 65535.0
        return temp, temp_c, temp_f, humidity

    def pack_data_structure(self, temp_f_lst, temp_c_lst, humidity_lst):
        """
        Calculate statistics and pack data structure.

        inputs:
            temp_f_lst(list): list of fehrenheit measurements
            temp_c_lst(list): list of celcius measurements
            humidity_lst(list): list of humidity measurements.
        returns:
            (dict): data structure.
        """
        return {sht31_config.API_MEASUREMENT_CNT: len(temp_f_lst),
                sht31_config.API_TEMPC_MEAN: statistics.mean(temp_c_lst),
                sht31_config.API_TEMPC_STD: statistics.pstdev(temp_c_lst),
                sht31_config.API_TEMPF_MEAN: statistics.mean(temp_f_lst),
                sht31_config.API_TEMPF_STD: statistics.pstdev(temp_f_lst),
                sht31_config.API_HUMIDITY_MEAN: statistics.mean(humidity_lst),
                sht31_config.API_HUMIDITY_STD: statistics.pstdev(humidity_lst),
                }

    def set_sht31_address(self, i2c_addr, addr_pin, alert_pin):
        """
        Set the address for the sht31.

        inputs:
            i2c_addr(int): bus address of SHT31.
            addr_pin(int):
            alert_pin(int):
        returns:
            None
        """
        # set address pin on SHT31
        GPIO.setmode(GPIO.BCM)  # broadcom pin numbering
        GPIO.setup(addr_pin, GPIO.OUT)  # address pin set as output
        GPIO.setup(alert_pin, GPIO.IN, pull_up_down=GPIO.PUD_UP)
        if i2c_addr == 0x45:
            GPIO.output(addr_pin, GPIO.HIGH)
        else:
            GPIO.output(addr_pin, GPIO.LOW)

    def send_i2c_cmd(self, bus, i2c_addr, i2c_command):
        """
        Send the i2c command to the sht31.

        inputs:
            bus(class 'SMBus'): i2c bus object
            i2c_addr(int):  bus address
            i2c_command(tuple): command (register, [data])
        returns:
            None
        """
        register = i2c_command[0]
        data = i2c_command[1]
        try:
            bus.write_i2c_block_data(i2c_addr, register, data)
        except OSError as exc:
            print(
                f"FATAL ERROR({util.get_function_name()}): i2c device at "
                f"address {hex(i2c_addr)} is not responding")
            raise exc
        time.sleep(0.5)

    def read_i2c_data(self, bus, i2c_addr, register=0x00, length=0x06):
        """
        Read i2c data.

        inputs:
            bus(class 'SMBus'): i2c bus object
            i2c_addr(int):  bus address
            register(int): register to read from.
            length(int):  number of blocks to read.
        returns:
            response(class 'list'): raw data structure
        """
        # Temp MSB, temp LSB, temp CRC, humidity MSB,
        # humidity LSB, humidity CRC
        # read_i2c_block_data(i2c_addr, register, length,
        #                     force=None)
        try:
            response = bus.read_i2c_block_data(i2c_addr,
                                               register,
                                               length)
        except OSError as exc:
            print(
                f"FATAL ERROR({util.get_function_name()}): i2c device at "
                f"address {hex(i2c_addr)} is not responding")
            raise exc
        return response

    def parse_fault_register_data(self, data):
        """
        Parse the fault register data, if possible.

        inputs:
            data(): list containing at least 2 bytes.
                    3rd byte (checksum) is not currently used.
        returns:
            (dict): fault register contents.
        """
        try:
            date_val = data[0] * 256 + data[1]
            resp = {
                'raw': data,
                'raw_binary': format(date_val, '#018b')[2:],
                'alert pending status(0=0,1=1+)': (date_val >> 15) & 1,
                # bit 15: 0=None, 1=at least 1 pending alert
                'heater status(0=off,1=on)': (date_val >> 13) & 1,
                # bit 13: 0=off, 1=on
                'RH tracking alert(0=no,1=yes)': (date_val >> 11) & 1,
                # bit 11: 0=no, 1=yes
                'T tracking alert(0=no,1=yes)': (date_val >> 10) & 1,
                # bit 10: 0=no, 1=yes
                'System reset detected(0=no,1=yes)': (date_val >> 4) & 1,
                # bit 4
                # 0=no reset since last clear cmd, 1=hard, soft, or supply fail
                'Command status(0=correct,1=failed)': (date_val >> 1) & 1,
                # bit 1: 0=successful, 1=invalid or field checksum
                'Write data checksum status(0=correct,1=failed)':
                    (date_val >> 0) & 1,
                # bit 0: 0=correct, 1=failed
            }
        except IndexError:
            # parsing error, just return raw data
            print("WARNING: fault register parsing error, just returning "
                  "raw fault register data")
            resp = {'raw': data}
        return resp

    def get_unit_test(self):
        """
        Get fabricated data for unit testing at ip:port/unit.

        syntax: http://ip:port/unit?seed=0x7E&measurements=1
        inputs:
            None
        returns:
            (dict): fabricated unit test data.
        """
        # get runtime parameters
        measurements = request.args.get('measurements',
                                        sht31_config.MEASUREMENTS,
                                        type=int)
        seed = request.args.get('seed', sht31_config.UNIT_TEST_SEED, type=int)

        # data structure
        temp_f_lst = []
        temp_c_lst = []
        humidity_lst = []

        # loop for n measurements
        for measurement in range(measurements):

            # fabricated data for unit testing
            data = [seed + measurement % 2] * 5  # almost mid range

            # convert the data
            _, temp_c, temp_f, humidity = self.convert_data(data)

            # add data to structure
            temp_f_lst.append(temp_f)
            temp_c_lst.append(temp_c)
            humidity_lst.append(humidity)

        # return data on API
        return self.pack_data_structure(temp_f_lst, temp_c_lst, humidity_lst)

    def get(self):
        """
        Get sensor data at ip:port.

        inputs:
            None
        returns:
            (dict): thermal data dictionary.
        """
        # get runtime parameters
        measurements = request.args.get('measurements', 1, type=int)

        # set address pin on SHT31
        self.set_sht31_address(sht31_config.I2C_ADDRESS, sht31_config.ADDR_PIN,
                               sht31_config.ALERT_PIN)

        # activate smbus
        bus = smbus.SMBus(sht31_config.I2C_BUS)
        time.sleep(0.5)

        # data structure
        temp_f_lst = []
        temp_c_lst = []
        humidity_lst = []

        try:
            # loop for n measurements
            for _ in range(measurements):
                # send single shot read command
                self.send_i2c_cmd(bus, sht31_config.I2C_ADDRESS,
                                  cs_enabled_high)

                # read the measurement data
                data = self.read_i2c_data(bus, sht31_config.I2C_ADDRESS,
                                          register=0x00, length=0x06)

                # convert the data
                _, temp_c, temp_f, humidity = self.convert_data(data)

                # add data to structure
                temp_f_lst.append(temp_f)
                temp_c_lst.append(temp_c)
                humidity_lst.append(humidity)

            # return data on API
            return self.pack_data_structure(temp_f_lst, temp_c_lst,
                                            humidity_lst)
        finally:
            # close the smbus connection
            bus.close()
            GPIO.cleanup()  # clean up GPIO

    def send_cmd_get_diag(self, i2c_command):
        """
        Send i2c command and read status register.

        inputs:
            i2c_command(int): i2c command to send
        returns:
            (dict): parsed fault register data.
        """
        # set address pin on SHT31
        self.set_sht31_address(sht31_config.I2C_ADDRESS, sht31_config.ADDR_PIN,
                               sht31_config.ALERT_PIN)

        # activate smbus
        bus = smbus.SMBus(sht31_config.I2C_BUS)
        time.sleep(0.5)

        try:
            # send single shot command
            self.send_i2c_cmd(bus, sht31_config.I2C_ADDRESS,
                              i2c_command)

            # small delay
            time.sleep(1.0)

            # read the measurement data, 2 bytes data, 1 byte checksum
            data = self.read_i2c_data(bus, sht31_config.I2C_ADDRESS,
                                      register=0x00, length=0x03)

            # parse data into registers
            parsed_data = self.parse_fault_register_data(data)
            return parsed_data
        finally:
            # close the smbus connection
            bus.close()
            GPIO.cleanup()  # clean up GPIO

    def i2c_recovery(self):
        """
        Send 10 clock cycles on SCL IO pin to clear locked i2c bus.

        inputs:
             None
        returns:
            (dict): status message
        """
        addr_pin = sht31_config.SCL_PIN
        num_clock_cycles = 10
        recovery_freq_hz = 100000  # 100 KHz
        recovery_delay_sec = 1.0 / (2.0 * recovery_freq_hz)

        try:
            # set SCL pin for output
            GPIO.setmode(GPIO.BCM)  # broadcom pin numbering
            GPIO.setup(addr_pin, GPIO.OUT)  # address pin set as output

            # set pin high to start
            GPIO.output(addr_pin, GPIO.HIGH)

            # send clock cycles
            for _ in range(num_clock_cycles):
                time.sleep(recovery_delay_sec)
                GPIO.output(addr_pin, GPIO.LOW)
                time.sleep(recovery_delay_sec)
                GPIO.output(addr_pin, GPIO.HIGH)
            # status message
            msg_dict = {}
            msg_dict["action_complete"] = (f"{num_clock_cycles} SCL clock "
                                           f"toggles completed at "
                                           f"{recovery_freq_hz} Hz")
            msg_dict["next_step"] = ("please reboot pi and restart "
                                     "flask server.")
            return {"i2c_recovery": msg_dict}
        finally:
            GPIO.cleanup()  # clean up GPIO

    def i2c_detect(self, bus=sht31_config.I2C_BUS):
        """
        Detect i2c device on bus.

        inputs:
             bus(int): i2c bus number
        returns:
            (dict): parsed device dictionary.
        """
        # send command
        p = subprocess.Popen(['sudo', 'i2cdetect', '-y',
                              str(bus)],
                             stdout=subprocess.PIPE,)
        # cmdout = str(p.communicate())

        # read in raw data
        parsed_device_dict = {"i2c_detect": {}}
        bus_dict = {}
        for _ in range(0, 9):
            line = str(p.stdout.readline())
            addr_base = line[2:4]
            addr_payload = line[5:]

            # catch error condition
            if "Error" in line:
                bus_dict["i2c_detect"]["error"] = line
            else:
                # find devices on bus
                device = 0
                device_dict = {}
                for match in re.finditer("[0-9][0-9]", addr_payload):
                    if match:
                        device_addr = match.group(0)
                        print(match.group(0))
                        device_dict["dev_" + str(device) + "_addr"] = \
                            str(device_addr)
                        bus_dict["addr_base_" + str(addr_base)] = device_dict
                        device += 1
        parsed_device_dict["i2c_detect"]["bus_" + str(bus)] = bus_dict
        return parsed_device_dict


class Controller(Resource):
    """Production controller."""

    def __init__(self):
        pass

    def get(self):
        """Map the get method."""
        helper = Sensors()
        return helper.get()


class ControllerUnit(Resource):
    """Unit test Controller."""

    def __init__(self):
        pass

    def get(self):
        """Map the get method."""
        helper = Sensors()
        return helper.get_unit_test()


class ReadFaultRegister(Resource):
    """Diagnostic Controller."""

    def __init__(self):
        pass

    def get(self):
        """Map the get method."""
        helper = Sensors()
        return helper.send_cmd_get_diag(read_status_register)


class ClearFaultRegister(Resource):
    """Clear Diagnostic Controller."""

    def __init__(self):
        pass

    def get(self):
        """Map the get method."""
        helper = Sensors()
        return helper.send_cmd_get_diag(clear_status_register)


class EnableHeater(Resource):
    """Enable heater Controller."""

    def __init__(self):
        pass

    def get(self):
        """Map the get method."""
        helper = Sensors()
        return helper.send_cmd_get_diag(enable_heater)


class DisableHeater(Resource):
    """Disable heater Controller."""

    def __init__(self):
        pass

    def get(self):
        """Map the get method."""
        helper = Sensors()
        return helper.send_cmd_get_diag(disable_heater)


class SoftReset(Resource):
    """i2C soft reset."""

    def __init__(self):
        pass

    def get(self):
        """Map the get method."""
        helper = Sensors()
        return helper.send_cmd_get_diag(soft_reset)


class Reset(Resource):
    """i2C hard reset."""

    def __init__(self):
        pass

    def get(self):
        """Map the get method."""
        helper = Sensors()
        return helper.send_cmd_get_diag(reset)


class I2CRecovery(Resource):
    """Issue i2c recovery sequence."""

    def __init__(self):
        pass

    def get(self):
        """Map the get method."""
        helper = Sensors()
        return helper.i2c_recovery()


class I2CDetect(Resource):
    """Issue i2c detect on default bus."""

    def __init__(self):
        pass

    def get(self):
        """Map the get method."""
        helper = Sensors()
        return helper.i2c_detect()


class I2CDetectBus0(Resource):
    """Issue i2c detect on bus 0."""

    def __init__(self):
        pass

    def get(self):
        """Map the get method."""
        helper = Sensors()
        return helper.i2c_detect(0)


class I2CDetectBus1(Resource):
    """Issue i2c detect on bus 1."""

    def __init__(self):
        pass

    def get(self):
        """Map the get method."""
        helper = Sensors()
        return helper.i2c_detect(1)


def create_app():
    """Create the api object."""
    app_ = Flask(__name__)

    # add API routes
    api = Api(app_)
    api.add_resource(Controller, "/")
    api.add_resource(ControllerUnit, sht31_config.flask_folder.unit_test)
    api.add_resource(ReadFaultRegister, sht31_config.flask_folder.diag)
    api.add_resource(ClearFaultRegister, sht31_config.flask_folder.clear_diag)
    api.add_resource(EnableHeater, sht31_config.flask_folder.enable_heater)
    api.add_resource(DisableHeater, sht31_config.flask_folder.disable_heater)
    api.add_resource(SoftReset, sht31_config.flask_folder.soft_reset)
    api.add_resource(Reset, sht31_config.flask_folder.reset)
    api.add_resource(I2CRecovery, sht31_config.flask_folder.i2c_recovery)
    api.add_resource(I2CDetect, sht31_config.flask_folder.i2c_detect)
    api.add_resource(I2CDetectBus0, sht31_config.flask_folder.i2c_detect_0)
    api.add_resource(I2CDetectBus1, sht31_config.flask_folder.i2c_detect_1)
    return app_


# create the flask app
app = create_app()
csrf = CSRFProtect(app)  # enable CSRF protection
ip_ban = flg.initialize_ipban(app)  # hacker blacklisting agent
flg.set_flask_cookie_config(app)
flg.print_flask_config(app)


@app.route('/favicon.ico')
def favicon():
    """Set favicon for browser tab."""
    return app.send_static_file('sht31.ico')


class UserInputs(util.UserInputs):
    """Manage runtime arguments for sht31_flask_server."""

    def __init__(self, argv_list=None, help_description=None,
                 suppress_warnings=False):
        """
        UserInputs constructor for sht31_flask_server.

        inputs:
            argv_list(list): override runtime values.
            help_description(str): description field for help text.
            suppress_warnings(bool): True to suppress warning msgs.
        """
        self.argv_list = argv_list
        self.help_description = help_description
        self.suppress_warnings = suppress_warnings

        # initialize parent class
        super().__init__(argv_list, help_description, suppress_warnings)

    def initialize_user_inputs(self, parent_keys=None):
        """
        Populate user_inputs dict.
        """
        if parent_keys is None:
            parent_keys = [self.default_parent_key]
        self.valid_sflags = []
        # define the user_inputs dict.
        for parent_key in parent_keys:
            self.user_inputs = {parent_key: {
                input_flds.debug_fld: {
                    "order": 1,    # index in the argv list
                    "value": None,
                    "type": lambda x: bool(str2bool(
                        str(x).strip())),
                    "default": False,
                    "valid_range": [True, False, 1, 0],
                    "sflag": "-d",
                    "lflag": "--" + input_flds.debug_fld,
                    "help": "flask server debug mode",
                    "required": False,
                    },
                },
            }
            self.valid_sflags += [self.user_inputs[parent_key][k]["sflag"]
                                  for k in self.user_inputs[parent_key].keys()]


if __name__ == "__main__":

    print("SHT31 sensor Flask server")

    # verify environment
    env.get_python_version()

    # parse runtime parameters
    uip = UserInputs()
    debug = uip.get_user_inputs(uip.default_parent_key, "debug")
    if debug:
        print("Flask debug mode is enabled", file=sys.stderr)

    # launch the Flask API on development server
    flg.schedule_ipban_block_list_report(ip_ban, debug_mode=debug)
    app.run(host='0.0.0.0',
            port=sht31_config.FLASK_PORT,
            debug=debug,
            threaded=True,  # threaded=True may speed up rendering on web page
            ssl_context=sht31_config.FLASK_SSL_CERT)



"""
Unit tes module for email_notification.py.
"""
# built-in libraries
import os
import smtplib
import unittest
from unittest import mock

# local libraries
from thermostatsupervisor import email_notification as eml
from thermostatsupervisor import environment as env
from thermostatsupervisor import utilities as util
from tests import unit_test_common as utc


class Test(utc.UnitTest):
    """Test email_notification.py functions."""

    to_address = None
    from_address = None
    from_password = None

    def test_check_email_env_variables(self):
        """
        Verify all required email email env variables are present for tests.

        If this test fails during AzDO CI, check repository secrets stored
        in Azure DevOps variables and also check yml file.
        If this test fails during manual run check env variables in
        local PC environment variables.
        """
        # make sure email account environmental variables are present
        for env_key in ['GMAIL_USERNAME', 'GMAIL_PASSWORD']:
            try:
                print(f"checking for environment variable key {env_key}")
                _ = os.environ[env_key]
                print(f"environment variable key {env_key} was found (PASS)")
            except KeyError:
                fail_msg = (f"{env_key} environment variable missing from "
                            f"environment")
                self.fail(fail_msg)

    def test_send_email_alerts(self):
        """Test send_email_alerts() functionality."""

        # send message with no inputs, UTIL.NO_ERROR expected
        body = "this is a test of the email notification alert."
        return_status, return_status_msg = \
            eml.send_email_alert(subject="test email alert",
                                 body=body)

        fail_msg = (f"send email with defaults failed for status code: "
                    f"{return_status}: {return_status_msg}")
        self.assertEqual(return_status, util.NO_ERROR, fail_msg)

        # send message with bad port, UTIL.CONNECTION_ERROR expected
        body = ("this is a test of the email notification alert with bad "
                "SMTP port input, should fail.")
        return_status, return_status_msg = \
            eml.send_email_alert(server_port=13,
                                 subject="test email alert "
                                 "(bad port)", body=body)
        fail_msg = (f"send email with bad server port failed for status code: "
                    f"{return_status}: {return_status_msg}")
        self.assertEqual(return_status, util.CONNECTION_ERROR, fail_msg)

        # send message with bad email address,
        # util.AUTHORIZATION_ERROR expected
        body = ("this is a test of the email notification alert with bad "
                "sender email address, should fail.")
        return_status, return_status_msg = \
            eml.send_email_alert(from_address="bogus@gmail.com",
                                 subject="test email alert "
                                 "(bad from address)", body=body)
        fail_msg = (f"send email with bad from addresss failed for status "
                    f"code: {return_status}: {return_status_msg}")
        self.assertEqual(return_status, util.AUTHORIZATION_ERROR, fail_msg)

    def test_send_email_alert_no_env_key(self):
        """Test send_email_alerts() functionality without email address."""

        # cache valid to_address
        to_address_valid = env.get_env_variable('GMAIL_USERNAME')

        # send message with no inputs, UTIL.NO_ERROR expected
        for name_to_omit in ["GMAIL_USERNAME", "GMAIL_PASSWORD"]:
            # to and from address currently share the same env key so add
            # runs with a valid to_address to confirm branch coverage of
            # a missing from address.
            for to_address in [None, to_address_valid]:
                valid_msg = ["valid", "invalid"][bool(to_address is None)]
                print(f"testing send_email_alert() with env key "
                      f"{name_to_omit} missing and {valid_msg} to_address...")
                modified_environ = utc.omit_env_vars(name_to_omit)

                body = (f"this is a test of the email notification alert with "
                        f"env var {name_to_omit} omitted.")
                with mock.patch.dict(os.environ, modified_environ, clear=True):  # noqa e501, pylint:disable=undefined-variable
                    return_status, return_status_msg = \
                        eml.send_email_alert(to_address=to_address,
                                             subject=f"test email alert with "
                                             f"env key {name_to_omit} missing",
                                             body=body)

                fail_msg = (f"send email with no email env key failed for "
                            f"status code: {return_status}: "
                            f"{return_status_msg}")
                self.assertEqual(return_status, util.ENVIRONMENT_ERROR,
                                 fail_msg)

    def test_send_email_alert_smtp_exceptions(self):
        """Test send_email_alerts() functionality with mocked exceptions."""
        bogus_exception_code = 99
        bogus_sender = "bogus sender"
        sendmail_exceptions = [(smtplib.SMTPHeloError,
                                [bogus_exception_code, "mock SMTPHeloError"]),
                               (smtplib.SMTPRecipientsRefused,
                                ["mock SMTPRecipientsRefused"]),
                               (smtplib.SMTPSenderRefused,
                                [bogus_exception_code,
                                 "mock SMTPSenderRefused", bogus_sender]),
                               (smtplib.SMTPDataError,
                                [bogus_exception_code, "mock SMTPDataError"]),
                               (smtplib.SMTPNotSupportedError,
                                ["mock SMTPNotSupportedError"]),
                               ]
        for exception, exception_args in sendmail_exceptions:
            print(f"testing mocked '{str(exception)} exception...")
            # mock the exception case
            side_effect = (lambda *_, **__: utc.mock_exception(exception,
                                                               exception_args))
            with mock.patch.object(smtplib.SMTP_SSL, 'sendmail',  # noqa e501, pylint:disable=undefined-variable
                                   side_effect=side_effect):
                # send message with no inputs, UTIL.NO_ERROR expected
                body = "this is a test of the email notification alert."
                return_status, return_status_msg = \
                    eml.send_email_alert(subject="test email alert",
                                         body=body)

            fail_msg = (f"send email with mocked SMTP exception returned "
                        f"status code: {return_status}: "
                        f"{return_status_msg}")
            self.assertEqual(return_status, util.EMAIL_SEND_ERROR,
                             fail_msg)


if __name__ == "__main__":
    util.log_msg.debug = True
    unittest.main(verbosity=2)



"""
Integration test module for mmm.py.

This test requires connection to mmm thermostat.
"""
# built-in imports
import unittest

# local imports
from thermostatsupervisor import mmm
from thermostatsupervisor import mmm_config
from thermostatsupervisor import utilities as util
from tests import unit_test_common as utc


class IntegrationTest(utc.IntegrationTest):
    """
    Test functions in mmm.py.
    """

    def setUpIntTest(self):
        """Setup common to integration tests."""
        self.setup_common()
        self.print_test_name()

        # argv list must be valid settings
        self.unit_test_argv = [
            "supervise.py",  # module
            "mmm50",  # thermostat
            "0",  # zone
            "30",  # poll time in sec
            "1000",  # reconnect time in sec
            "2",  # tolerance
            "UNKNOWN_MODE",  # thermostat mode, no target
            "3",  # number of measurements
        ]
        self.mod = mmm
        self.mod_config = mmm_config


@unittest.skipIf(not utc.ENABLE_MMM_TESTS,
                 "mmm tests are disabled")
class FunctionalIntegrationTest(IntegrationTest,
                                utc.FunctionalIntegrationTest):
    """
    Test functional performance of mmm.py.
    """

    def setUp(self):
        super().setUp()
        self.setUpIntTest()
        # test_GetMetaData input parameters
        self.metadata_field = "network"
        self.metadata_type = dict

    def test_thermostat_get_ui_data(self):
        """Verify Thermostat.get_ui_data() function"""
        # setup class instances
        self.Thermostat, self.Zone = self.setup_thermostat_zone()

        # verify function return value
        result = self.Thermostat.get_ui_data(self.Thermostat.zone_name)
        print(f"Thermostat.get_ui_data returned {result}")
        self.assertTrue(isinstance(result, dict),
                        f"result returned is type ({type(result)}), "
                        "expected a dictionary")

    def test_thermostat_get_ui_data_param(self):
        """Verify Thermostat.get_ui_data_param() function"""
        # setup class instances
        self.Thermostat, self.Zone = self.setup_thermostat_zone()

        # verify function return value
        param = "cloud"
        expected_data_type = dict
        result = self.Thermostat.get_ui_data_param(self.Thermostat.zone_name,
                                                   param)
        print(f"Thermostat.get_ui_data_param({self.Thermostat.zone_name}, "
              f" {param}) returned {result}")
        self.assertTrue(isinstance(result, expected_data_type),
                        f"result returned is type ({type(result)}), expected "
                        f"a {expected_data_type}")

    def test_zone_get_zone_name(self):
        """Verify Zone.get_zone_name() function"""
        # setup class instances
        self.Thermostat, self.Zone = self.setup_thermostat_zone()

        # verify default option
        result = self.Zone.get_zone_name()
        print(f"Zone.get_zone_name() returned {result}")

        # verify override option
        result = self.Zone.get_zone_name(self.Thermostat.zone_name)
        print(f"Zone.get_zone_name({self.Thermostat.zone_name}) "
              f"returned {result}")


@unittest.skipIf(not utc.ENABLE_MMM_TESTS,
                 "mmm tests are disabled")
class SuperviseIntegrationTest(IntegrationTest,
                               utc.SuperviseIntegrationTest):
    """
    Test supervise functionality of mmm.py.
    """

    def setUp(self):
        super().setUp()
        self.setUpIntTest()


@unittest.skipIf(not utc.ENABLE_MMM_TESTS,
                 "mmm tests are disabled")
class PerformanceIntegrationTest(IntegrationTest,
                                 utc.PerformanceIntegrationTest):
    """
    Test performance of in mmm.py.
    """

    def setUp(self):
        super().setUp()
        self.setUpIntTest()
        # network timing measurement
        self.timeout_limit = mmm.SOCKET_TIMEOUT
        self.timing_measurements = 100


if __name__ == "__main__":
    util.log_msg.debug = True
    unittest.main(verbosity=2)



"""emulator integration"""
import random
import traceback

# local imports
from thermostatsupervisor import emulator_config
from thermostatsupervisor import environment as env
from thermostatsupervisor import thermostat_api as api
from thermostatsupervisor import thermostat_common as tc
from thermostatsupervisor import utilities as util


class ThermostatClass(tc.ThermostatCommon):
    """Emulator thermostat functions."""

    def __init__(self, zone, verbose=True):
        """
        Constructor, connect to thermostat.

        inputs:
            zone(str):  zone of thermostat.
            verbose(bool): debug flag.
        """
        # construct the superclass
        # call both parent class __init__
        tc.ThermostatCommon.__init__(self)

        # set tstat type and debug flag
        self.thermostat_type = emulator_config.ALIAS
        self.verbose = verbose

        # configure zone info
        self.zone_name = int(zone)
        self.device_id = self.get_target_zone_id(self.zone_name)
        self.serial_number = None  # will be populated when unit is queried.
        self.meta_data_dict = {}
        self.initialize_meta_data_dict()

    def initialize_meta_data_dict(self):
        """Initialize the meta data dict"""
        # add zone keys
        for key in emulator_config.supported_configs["zones"]:
            self.meta_data_dict[key] = {}

    def get_target_zone_id(self, zone=0):
        """
        Return the target zone ID.

        inputs:
            zone(int):  zone number.
        returns:
            (int): device_id
        """
        return zone

    def get_all_metadata(self, zone=None, debug=False):
        """Get all thermostat meta data for zone from kumocloud.

        inputs:
            zone(int): specified zone, if None will print all zones.
            debug(bool): if True will print unit details.
        returns:
            (dict): JSON dict
        """
        return self.get_metadata(zone, None, debug)

    def get_metadata(self, zone=None, parameter=None, debug=False):
        """Get all thermostat meta data for zone from emulator.

        inputs:
            zone(int): specified zone, if None will print all zones.
            parameter(str): target parameter, if None will return all.
            debug(bool): if True will print unit details.
        returns:
            (int, float, str, dict): depends on parameter
        """
        del debug  # unused
        if zone is None:
            # returned cached raw data for all zones
            meta_data_dict = self.meta_data_dict
        else:
            # return cached raw data for specified zone
            meta_data_dict = self.meta_data_dict[zone]
        if parameter is None:
            return meta_data_dict
        else:
            return meta_data_dict[parameter]

    def print_all_thermostat_metadata(self, zone, debug=False):
        """Print all metadata for zone to the screen.

        inputs:
            zone(int): specified zone, if None will print all zones.
            debug(bool): debug flag
        returns:
            None, prints result to screen
        """
        self.exec_print_all_thermostat_metadata(
            self.get_all_metadata, [zone, debug])


class ThermostatZone(tc.ThermostatCommonZone):
    """
    KumoCloud single zone from kumocloud.

    Class needs to be updated for multi-zone support.
    """

    def __init__(self, Thermostat_obj, verbose=True):
        """
        Zone constructor.

        inputs:
            Thermostat(obj): Thermostat class instance.
            verbose(bool): debug flag.
        """
        # construct the superclass, requires auth setup first
        super().__init__()

        # runtime parameter defaults
        self.poll_time_sec = 1 * 60  # default to 1 minutes
        self.connection_time_sec = 1 * 60 * 60  # default to 1 hours

        # switch config for this thermostat, numbers are unique and arbitrary
        self.system_switch_position[
            tc.ThermostatCommonZone.OFF_MODE] = 0
        self.system_switch_position[
            tc.ThermostatCommonZone.HEAT_MODE] = 1
        self.system_switch_position[
            tc.ThermostatCommonZone.COOL_MODE] = 2
        self.system_switch_position[
            tc.ThermostatCommonZone.DRY_MODE] = 3
        self.system_switch_position[
            tc.ThermostatCommonZone.AUTO_MODE] = 4
        self.system_switch_position[
            tc.ThermostatCommonZone.FAN_MODE] = 5

        # zone info
        self.verbose = verbose
        self.thermostat_type = emulator_config.ALIAS
        self.device_id = Thermostat_obj.device_id
        self.Thermostat = Thermostat_obj
        self.zone_name = Thermostat_obj.zone_name
        self.zone_info = Thermostat_obj.get_all_metadata(
            Thermostat_obj.zone_name)
        self.zone_name = self.get_zone_name()
        self.initialize_meta_data_dict()

    def initialize_meta_data_dict(self):
        """Initialize the meta data dict"""
        # add parameters and values
        self.set_heat_setpoint(emulator_config.STARTING_TEMP)
        self.set_cool_setpoint(emulator_config.STARTING_TEMP)
        self.set_parameter('display_temp', emulator_config.STARTING_TEMP)
        self.set_parameter('display_humidity',
                           emulator_config.STARTING_HUMIDITY)
        self.set_parameter('humidity_support', True)
        self.set_parameter('power_on', True)
        self.set_parameter('fan_on', True)
        self.set_parameter('fan_speed', 3)
        self.set_parameter('defrost', False)
        self.set_parameter('standby', False)
        self.set_parameter('vacation_hold', False)
        self.set_mode(emulator_config.STARTING_MODE)

    def get_parameter(self, key,
                      default_val=None):
        """
        Get parameter from zone dictionary.

        inputs:
            key(str): target dict key
            default_val(str, int, float): default value on key errors
        """
        return_val = default_val
        try:
            return_val = self.zone_info[key]
        except KeyError:
            util.log_msg(traceback.format_exc(),
                         mode=util.BOTH_LOG, func_name=1)
        return return_val

    def set_parameter(self, key, target_val=None):
        """
        Set parameter in zone dictionary.

        inputs:
            key(str): target dict key
            target_val(str, int, float): value to set
        """
        self.zone_info[key] = target_val

    def get_zone_name(self):
        """
        Return the name associated with the zone number.

        inputs:
            None
        returns:
            (str) zone name
        """
        self.refresh_zone_info()
        return "emulator_" + str(self.zone_name)

    def get_display_temp(self) -> float:  # used
        """
        Refresh the cached zone information and return Indoor Temp in Deg F
        with +/- 1 degree noise value.

        inputs:
            None
        returns:
            (float): indoor temp in deg F.
        """
        self.refresh_zone_info()
        return (self.get_parameter('display_temp') +
                random.uniform(-emulator_config.NORMAL_TEMP_VARIATION,
                               emulator_config.NORMAL_TEMP_VARIATION))

    def get_display_humidity(self) -> (float, None):
        """
        Refresh the cached zone information and return IndoorHumidity
        with random +/-1% noise value.

        inputs:
            None
        returns:
            (float, None): indoor humidity in %RH, None if not supported.
        """
        if not self.get_is_humidity_supported():
            return None
        else:
            return (self.get_parameter('display_humidity') +
                    random.uniform(-emulator_config.NORMAL_HUMIDITY_VARIATION,
                                   emulator_config.NORMAL_HUMIDITY_VARIATION))

    def get_is_humidity_supported(self) -> bool:  # used
        """
        Refresh the cached zone information and return the
          True if humidity sensor data is trustworthy.

        inputs:
            None
        returns:
            (booL): True if is in humidity sensor is available and not faulted.
        """
        self.refresh_zone_info()
        return self.get_parameter('humidity_support')

    def set_mode(self, target_mode):
        """
        Set the thermostat mode.

        inputs:
            target_mode(str): target mode, refer to supported_configs["modes"]
        returns:
            True if successful, else False
        """
        if self.verbose:
            print(f"setting mode to {target_mode}")
        self.set_parameter('switch_position',
                           self.system_switch_position[
                               getattr(tc.ThermostatCommonZone, target_mode)])

    def is_heat_mode(self) -> int:
        """
        Refresh the cached zone information and return the heat mode.

        inputs:
            None
        returns:
            (int) heat mode, 1=enabled, 0=disabled.
        """
        return int(self.get_system_switch_position() ==
                   self.system_switch_position[
                       tc.ThermostatCommonZone.HEAT_MODE])

    def is_cool_mode(self) -> int:
        """
        Refresh the cached zone information and return the cool mode.

        inputs:
            None
        returns:
            (int): cool mode, 1=enabled, 0=disabled.
        """
        return int(self.get_system_switch_position() ==
                   self.system_switch_position[
                       tc.ThermostatCommonZone.COOL_MODE])

    def is_dry_mode(self) -> int:
        """
        Refresh the cached zone information and return the dry mode.

        inputs:
            None
        returns:
            (int): dry mode, 1=enabled, 0=disabled.
        """
        return int(self.get_system_switch_position() ==
                   self.system_switch_position[
                       tc.ThermostatCommonZone.DRY_MODE])

    def is_fan_mode(self) -> int:
        """
        Refresh the cached zone information and return the fan mode.

        inputs:
            None
        returns:
            (int): fan mode, 1=enabled, 0=disabled.
        """
        return int(self.get_system_switch_position() ==
                   self.system_switch_position[
                       tc.ThermostatCommonZone.FAN_MODE])

    def is_auto_mode(self) -> int:
        """
        Refresh the cached zone information and return the auto mode.

        inputs:
            None
        returns:
            (int): auto mode, 1=enabled, 0=disabled.
        """
        return int(self.get_system_switch_position() ==
                   self.system_switch_position[
                       tc.ThermostatCommonZone.AUTO_MODE])

    def is_off_mode(self) -> int:
        """
        Refresh the cached zone information and return the off mode.

        inputs:
            None
        returns:
            (int): off mode, 1=enabled, 0=disabled.
        """
        return int(self.get_system_switch_position() ==
                   self.system_switch_position[
                       tc.ThermostatCommonZone.OFF_MODE])

    def is_heating(self):
        """Return 1 if heating relay is active, else 0."""
        return int(self.is_heat_mode() and self.is_power_on() and
                   self.get_heat_setpoint_raw() > self.get_display_temp())

    def is_cooling(self):
        """Return 1 if cooling relay is active, else 0."""
        return int(self.is_cool_mode() and self.is_power_on() and
                   self.get_cool_setpoint_raw() < self.get_display_temp())

    def is_drying(self):
        """Return 1 if drying relay is active, else 0."""
        return int(self.is_dry_mode() and self.is_power_on() and
                   self.get_cool_setpoint_raw() < self.get_display_temp())

    def is_auto(self):
        """Return 1 if auto relay is active, else 0."""
        return int(self.is_auto_mode() and self.is_power_on() and
                   (self.get_cool_setpoint_raw() < self.get_display_temp() or
                    self.get_heat_setpoint_raw() > self.get_display_temp()))

    def is_fanning(self):
        """Return 1 if fan relay is active, else 0."""
        return int(self.is_fan_on() and self.is_power_on())

    def is_power_on(self):
        """Return 1 if power relay is active, else 0."""
        self.refresh_zone_info()
        return self.get_parameter('power_on')

    def is_fan_on(self):
        """Return 1 if fan relay is active, else 0."""
        self.refresh_zone_info()
        return self.get_parameter('fan_speed') > 0

    def is_defrosting(self):
        """Return 1 if defrosting is active, else 0."""
        self.refresh_zone_info()
        return int(self.get_parameter('defrost'))

    def is_standby(self):
        """Return 1 if standby is active, else 0."""
        self.refresh_zone_info()
        return int(self.get_parameter('standby'))

    def get_heat_setpoint_raw(self) -> int:  # used
        """
        Refresh the cached zone information and return the heat setpoint.

        inputs:
            None
        returns:
            (int): heating set point in degrees F.
        """
        self.refresh_zone_info()
        return self.get_parameter('heat_setpoint')

    def get_heat_setpoint(self) -> str:
        """Return heat setpoint with units as a string."""
        return util.temp_value_with_units(self.get_heat_setpoint_raw())

    def get_schedule_heat_sp(self) -> int:  # used
        """
        Return the schedule heat setpoint.

        inputs:
            None
        returns:
            (int): scheduled heating set point in degrees.
        """
        return emulator_config.MAX_HEAT_SETPOINT  # max heat set point allowed

    def get_schedule_cool_sp(self) -> int:
        """
        Return the schedule cool setpoint.

        inputs:
            None
        returns:
            (int): scheduled cooling set point in degrees F.
        """
        return emulator_config.MIN_COOL_SETPOINT  # min cool set point allowed

    def get_cool_setpoint_raw(self) -> int:
        """
        Return the cool setpoint.

        inputs:
            None
        returns:
            (int): cooling set point in degrees F.
        """
        self.refresh_zone_info()
        return self.get_parameter('cool_setpoint')

    def get_cool_setpoint(self) -> str:
        """Return cool setpoint with units as a string."""
        return util.temp_value_with_units(self.get_cool_setpoint_raw())

    def get_is_invacation_hold_mode(self) -> bool:  # used
        """
        Return the
          'IsInVacationHoldMode' setting.

        inputs:
            None
        returns:
            (booL): True if is in vacation hold mode.
        """
        return self.get_parameter('vacation_hold')

    def get_vacation_hold(self) -> bool:
        """
        Return the
        VacationHold setting.

        inputs:
            None
        returns:
            (bool): True if vacation hold is set.
        """
        # TODO, are vacationhold unique fields?  what used for?
        return self.get_parameter('vacation_hold')

    def get_system_switch_position(self) -> int:  # used
        """
        Return the system switch position.

        inputs:
            None
        returns:
            (int) current mode for unit, should match value
                  in self.system_switch_position
        """
        self.refresh_zone_info()
        # first check if power is on
        # if power is off then operation_mode key may be missing.
        if not self.is_power_on():
            return self.system_switch_position[
                tc.ThermostatCommonZone.OFF_MODE]
        else:
            return self.get_parameter('switch_position')

    def set_heat_setpoint(self, temp: int) -> None:
        """
        Set a new heat setpoint.

        inputs:
            temp(int): desired temperature in F
        returns:
            None
        """
        self.set_parameter('heat_setpoint', temp)

    def set_cool_setpoint(self, temp: int) -> None:
        """
        Set a new cool setpoint.

        inputs:
            temp(int): desired temperature in deg F.
        returns:
            None
        """
        self.set_parameter('cool_setpoint', temp)

    def refresh_zone_info(self, force_refresh=False):
        """
        Refresh zone info from KumoCloud.

        inputs:
            force_refresh(bool): if True, ignore expiration timer.
        returns:
            None, zone_data is refreshed.
        """
        del force_refresh
        # do nothing

    def report_heating_parameters(self, switch_position=None):
        """
        Display critical thermostat settings and reading to the screen.

        inputs:
            switch_position(int): switch position override, used for testing.
        returns:
            None
        """
        # current temp as measured by thermostat
        util.log_msg(f"display temp="
                     f"{util.temp_value_with_units(self.get_display_temp())}",
                     mode=util.BOTH_LOG,
                     func_name=1)

        # get switch position
        if switch_position is None:
            switch_position = self.get_system_switch_position()

        # heating status
        if switch_position == \
                self.system_switch_position[self.HEAT_MODE]:
            util.log_msg(f"heat mode={self.is_heat_mode()}",
                         mode=util.BOTH_LOG)
            util.log_msg(
                f"heat setpoint={self.get_heat_setpoint_raw()}",
                mode=util.BOTH_LOG)
            util.log_msg(
                f"schedule heat sp={self.get_schedule_heat_sp()}",
                mode=util.BOTH_LOG)

        # cooling status
        if switch_position == \
                self.system_switch_position[self.COOL_MODE]:
            util.log_msg(f"cool mode={self.is_cool_mode()}",
                         mode=util.BOTH_LOG)
            util.log_msg(
                f"cool setpoint={self.get_cool_setpoint_raw()}",
                mode=util.BOTH_LOG)
            util.log_msg(
                f"schedule cool sp={self.get_schedule_cool_sp()}",
                mode=util.BOTH_LOG)

        # hold settings
        util.log_msg(
            f"is in vacation hold mode={self.get_is_invacation_hold_mode()}",
            mode=util.BOTH_LOG)
        util.log_msg(f"vacation hold={self.get_vacation_hold()}",
                     mode=util.BOTH_LOG)
        util.log_msg(
            f"vacation hold until time={self.get_vacation_hold_until_time()}",
            mode=util.BOTH_LOG)
        util.log_msg(f"temporary hold until time="
                     f"{self.get_temporary_hold_until_time()}",
                     mode=util.BOTH_LOG)


if __name__ == "__main__":

    # verify environment
    env.get_python_version()

    # get zone override
    api.uip = api.UserInputs(argv_list=None,
                             thermostat_type=emulator_config.ALIAS)
    zone_number = api.uip.get_user_inputs(api.uip.zone_name,
                                          api.input_flds.zone)

    tc.thermostat_basic_checkout(
        emulator_config.ALIAS,
        zone_number,
        ThermostatClass, ThermostatZone)

    tc.thermostat_get_all_zone_temps(
        emulator_config.ALIAS,
        emulator_config.supported_configs["zones"],
        ThermostatClass,
        ThermostatZone)



"""
Thermostat Supervisor
"""
# built ins
import sys
import time

# local imports
from thermostatsupervisor import environment as env
from thermostatsupervisor import thermostat_api as api
from thermostatsupervisor import utilities as util

argv = []  # runtime parameter override


def supervisor(thermostat_type, zone_str):
    """
    Monitor specified thermometer and zone for deviations up to max
    measurements.

    inputs:
        thermostat_type(str): thermostat type, see thermostat_api for list
                              of supported thermostats.
        zone_str(str):        zone number input from user
    returns:
        None
    """
    # session variables:
    debug = False  # verbose debugging information

    # load hardware library
    mod = api.load_hardware_library(thermostat_type)

    # verify env variables are present
    api.verify_required_env_variables(thermostat_type, zone_str)

    # connection timer loop
    session_count = 1
    measurement = 1

    # outer loop: sessions
    while not api.uip.max_measurement_count_exceeded(measurement):
        # make connection to thermostat
        zone_num = api.uip.get_user_inputs(api.uip.zone_name,
                                           api.input_flds.zone)
        util.log_msg(
            f"connecting to thermostat zone {zone_num} "
            f"(session:{session_count})...",
            mode=util.BOTH_LOG)
        Thermostat = mod.ThermostatClass(zone_num)

        # dump all meta data
        if debug:
            util.log_msg("thermostat meta data:", mode=util.BOTH_LOG,
                         func_name=1)
            Thermostat.print_all_thermostat_metadata(zone_num)

        # get Zone object based on deviceID
        Zone = mod.ThermostatZone(Thermostat)
        util.log_msg(f"zone name={Zone.zone_name}", mode=util.BOTH_LOG,
                     func_name=1)

        # display banner and session settings
        Zone.display_session_settings()

        # set start time for poll
        Zone.session_start_time_sec = time.time()

        # update runtime overrides
        Zone.update_runtime_parameters()

        # display runtime settings
        Zone.display_runtime_settings()

        # supervisor inner loop
        measurement = Zone.supervisor_loop(Thermostat, session_count,
                                           measurement, debug)

        # increment connection count
        session_count += 1

    # clean-up and exit
    util.log_msg(
        f"\n{measurement - 1} measurements completed, exiting program\n",
        mode=util.BOTH_LOG)

    # delete packages if necessary
    if 'Zone' in locals():
        del Zone
    if 'Thermostat' in locals():
        del Thermostat
    if 'mod' in locals():
        del mod


def exec_supervise(debug=True, argv_list=None):
    """
    Execute supervisor loop.

    inputs:
        debug(bool): enable debugging mode.
        argv_list(list): argv overrides.
    returns:
        (bool): True if complete.
    """
    util.log_msg.debug = debug  # debug mode set

    # parse all runtime parameters if necessary
    api.uip = api.UserInputs(argv_list)

    # main supervise function
    # TODO - update for multi-zone application
    supervisor(api.uip.get_user_inputs(api.uip.parent_keys[0],
                                       api.input_flds.thermostat_type),
               api.uip.get_user_inputs(api.uip.parent_keys[0],
                                       api.input_flds.zone)
               )

    return True


if __name__ == "__main__":

    # if argv list is set use that, else use sys.argv
    if argv:
        argv_inputs = argv
    else:
        argv_inputs = sys.argv

    # verify environment
    env.get_python_version()

    exec_supervise(debug=True, argv_list=argv_inputs)
